# DDEP database, Decay Data Evaluation Project

import json
import hdtv.util

try:
    import urllib.request
    import urllib.error
except ImportError:
    import urllib
import hdtv.ui
from uncertainties import ufloat
from hdtv.database.common import *


def SearchNuclide(nuclide):
    """
    Opens table of nuclides with peak energies, gives back the peak energies of the nuclide and its intensities.
    """

    out = {"nuclide": nuclide, "transitions": []}

    if nuclide == "Ra-226":
        nuclide = "Ra-226D"

    try:
        with urllib.request.urlopen(
            "http://www.nucleide.org/DDEP_WG/Nuclides/" + str(nuclide) + ".lara.txt"
        ) as resource:
            data = resource.read().decode("utf-8")
    except:
        raise hdtv.cmdline.HDTVCommandError(f"Error looking up nuclide {nuclide}")

    for line in data.split("\r\n"):
        sep = line.split(" ; ")

        if str(sep[0]) == "Half-life (s)":
            out["halflife"] = ufloat(float(sep[1]), float(sep[2]))

        if str(sep[0]) == "Reference":
            out["reference"] = str(sep[1])

        try:
            if str(sep[4]) == "g":
                energy = ufloat(float(sep[0]), 0)
                try:
                    energy.std_dev = float(sep[1])
                except BaseException:
                    pass

                intensity = ufloat(float(sep[2]) / 100, 0)
                try:
                    intensity.std_dev = float(sep[3]) / 100
                except BaseException:
                    pass

                out["transitions"].append({"energy": energy, "intensity": intensity})
        except:
            pass

    return out



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2019  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Basic user interface functions (Input/Output, etc)
"""

import sys
import os
from html import escape
from typing import cast, TextIO

import asyncio

import hdtv.options
import hdtv.cmdline

from prompt_toolkit.patch_stdout import StdoutProxy, patch_stdout
from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import HTML, ANSI


class SimpleUI:
    """
    Very simple UserInterface
    """

    def __init__(self):
        self.stdout = None
        self.stderr = None

    def print(self, html, end="\n", err=False):
        # Currently, patch_stdout and print_formatted_text don’t work
        # together. Therefore, we clear the current line manually
        # first, print the formatted text, and finally call print to
        # force a redraw of the prompt. This causes problems when the
        # prompt is multiple lines long (the prompt will overwrite
        # the last lines that were just drawn).
        if self.stderr and err:
            self.stderr.write(html)
            return
        if self.stdout:
            self.stdout.write(html)
            return
        sys.stdout.write("\r")
        # asyncio.set_event_loop(hdtv.cmdline.command_line.loop)
        # self.print_patched(HTML(html), end)
        hdtv.cmdline.command_line.loop.call_soon_threadsafe(
            self.print_patched, HTML(html), end
        )

    def print_patched(self, text, end):
        with patch_stdout(raw=False):
            print_formatted_text(text, end=end)
            print("", end="\r")

    def msg(self, text=None, *, html=None, end="\n"):
        """
        Print message
        """
        if not html:
            html = escape(text or "")
        self.print(html, end=end)

    def info(self, text=None, *, html=None, end="\n"):
        """
        Print informational message
        """
        if not html:
            html = escape(text)
        html = "INFO: " + html
        self.print(text, end=end)

    def warning(self, text=None, *, html=None, end="\n"):
        """
        Print warning message
        """
        if not html:
            html = escape(text)
        self.print(f"<ansiyellow>WARNING: {html}</ansiyellow>", end=end, err=True)

    def error(self, text=None, *, html=None, end="\n"):
        """
        Print error message
        """
        if not html:
            html = escape(text)
        self.print(f"<ansired>ERROR: {html}</ansired>", end=end, err=True)

    def debug(self, text=None, *, html=None, end="\n", level=1):
        """
        Debugging output. The higher the level, the more specific is
        the debug message.
        """
        if not html:
            html = escape(text)
        self.print(f"<ansiblue>DEBUG: {html}</ansiblue>", end=end, err=True)


# Initialization
ui = SimpleUI()

msg = ui.msg
info = ui.info
warning = ui.warning
error = ui.error


def debug(text, end="\n", level=1):
    if level > hdtv.options.Get("ui.out.level"):
        return
    else:
        ui.debug(text, end=end, level=level)


opt = hdtv.options.Option(default=0, parse=lambda x: int(x))
hdtv.options.RegisterOption("ui.out.level", opt)



#!/usr/bin/env python3

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

# -------------------------------------------------------------------------
# Test program to compare fitting results of tv and hdtv
# -------------------------------------------------------------------------

import peakgen

spec = peakgen.Spectrum(1024, -0.5, 1023.5)
spec.func.background = peakgen.PolyBg([10.0])
# spec.func.peaks.append(peakgen.TheuerkaufPeak(500, 300.0, 10.0, sh=0.3, sw=1.0))
spec.func.peaks.append(peakgen.TheuerkaufPeak(515, 3000.0, 10.0))
# spec.peaks.append(peakgen.EEPeak(300, 100, 4.5, 6.0, 1.5, 0.7))
# spec.peaks.append(peakgen.EEPeak(400, 30, 4.5, 6.0, 1.5, 0.7))

nsamples = int(3e3)
hs = spec.GetSampledHist(nsamples)
# hs.Draw()

he = spec.GetExactHist()
scale = nsamples / he.GetSumOfWeights()
he.Scale(scale)

# he.SetLineColor(ROOT.kRed)
# he.Draw("SAME")

peakgen.write_hist(he, "test_exact.asc")
peakgen.write_hist(hs, "test_sampled.asc")



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2019  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
main session of hdtv
"""

import copy

import hdtv.cal

from hdtv.window import Window
from hdtv.drawable import DrawableManager
from hdtv.fitter import Fitter
from hdtv.fit import Fit
from hdtv.cut import Cut
from hdtv.integral import Integrate
from hdtv.util import monkey_patch_ui


class Session(DrawableManager):
    """
    Main session of hdtv

    First of all this provides a list of spectra, which is why this is called
    spectra in most contexts. But this also keeps track of the basic fit interface
    and of a list of calibrations.
    """

    def __init__(self):
        self.window = Window()
        super().__init__(viewport=self.window.viewport)
        # TODO: make peakModel and bgdeg configurable
        self.workFit = Fit(Fitter(peakModel="theuerkauf", backgroundModel="polynomial"))
        self.workFit.active = True
        self.workFit.Draw(self.viewport)
        self.workCut = Cut()
        self.workCut.active = True
        self.workCut.Draw(self.viewport)
        self.caldict = hdtv.cal.PositionCalibrationDict()
        # main session is always active
        self._active = True

    def ApplyCalibration(self, specIDs, cal, storeInCaldict: bool = True):
        """
        Apply calibration cal to spectra with ids
        """
        if isinstance(specIDs, (str, int, float)):
            specIDs = hdtv.util.ID.ParseIds(str(specIDs), self)
        for ID in specIDs:
            try:
                spec = self.dict[ID]
            except KeyError:
                hdtv.ui.warning("There is no spectrum with id: %s" % ID)
            else:
                if cal is None:
                    hdtv.ui.msg("Unsetting calibration of spectrum with id %s" % ID)
                    try:
                        self.caldict.pop(spec.name)
                    except KeyError:
                        pass
                    spec.cal = None
                else:
                    hdtv.ui.msg("Calibrated spectrum with id %s" % ID)
                    cal = hdtv.cal.MakeCalibration(cal)
                    if storeInCaldict:
                        self.caldict[spec.name] = cal
                    spec.cal = cal
                if self.workFit.spec == spec:
                    self.workFit.cal = cal

    # Marker handling
    def SetMarker(self, mtype, pos=None):
        """
        Set Marker of type "mtype" (bg, peak, region, cut) to position pos,
        if pos is not given, the current position of the cursor is used.
        """
        if pos is None:
            pos = self.viewport.GetCursorX()
        if mtype in ["bg", "peak", "region"]:
            self.workFit.ChangeMarker(mtype, pos, action="set")
        elif mtype in ["cut", "cutregion", "cutbg"]:
            mtype = mtype[3:]
            self.workCut.SetMarker(mtype, pos)

    def RemoveMarker(self, mtype, pos=None):
        """
        Remove the marker of type "mtype" (bg, peak, region), that is
        closest to position "pos", if position is not given, the current position
        of the cursor is used.
        """
        if pos is None:
            pos = self.viewport.GetCursorX()
        if mtype in ["bg", "peak", "region"]:
            self.workFit.ChangeMarker(mtype, pos, action="remove")
        elif mtype in ["cutregion", "cutbg"]:
            mtype = mtype[3:]
            self.workCut.RemoveMarker(mtype, pos)

    def ExecuteIntegral(self):
        spec = self.GetActiveObject()
        if spec is None:
            hdtv.ui.error("There is no active spectrum.")
            return

        fit = self.workFit
        fit.spec = spec
        if not fit.regionMarkers.IsFull():
            hdtv.ui.error("Region not set.")
            return

        if len(fit.bgMarkers) > 0:
            if fit.fitter.backgroundModel.fParStatus["nparams"] == -1:
                hdtv.ui.error("Background degree of -1 contradicts background fit.")
                return
            # pure background fit
            fit.FitBgFunc(spec)

        region = [fit.regionMarkers[0].p1.pos_uncal, fit.regionMarkers[0].p2.pos_uncal]
        bg = fit.fitter.bgFitter

        fit.integral = Integrate(spec, bg, region)
        fit.Draw(self.viewport)
        hdtv.ui.msg(html=fit.print_integral())

    # Functions to handle workFit
    def ExecuteFit(self, peaks=True):
        """
        Execute the fit

        If peaks=False, just a background fit is executed, else a peak fit is executed.
        """
        hdtv.ui.debug("Executing the fit")
        spec = self.GetActiveObject()
        if spec is None:
            hdtv.ui.error("There is no active spectrum.")
            return

        fit = self.workFit
        try:
            if not peaks and len(fit.bgMarkers) > 0:
                fit.FitBgFunc(spec)
            if peaks:
                # full fit
                fit.FitPeakFunc(spec)
            # show fit result
            hdtv.ui.msg(html=str(fit), end="")
            fit.Draw(self.viewport)
        except OverflowError as msg:
            hdtv.ui.error("Fit failed: %s" % msg)
        if fit.regionMarkers.IsFull():
            region = [
                fit.regionMarkers[0].p1.pos_uncal,
                fit.regionMarkers[0].p2.pos_uncal,
            ]
            fit.integral = Integrate(spec, fit.fitter.bgFitter, region)

    def ClearFit(self, bg_only=False):
        """
        Clear the markers and erase a previously executed fit.

        If bg_only is True, only the background fit is cleared and a peak fit
        is repeated now with internal background.
        """
        if bg_only:
            self.workFit.bgMarkers.Clear()
            self.workFit.Erase(bg_only)
            self.workFit.Refresh()
        else:
            self.workFit.bgMarkers.Clear()
            self.workFit.peakMarkers.Clear()
            self.workFit.regionMarkers.Clear()
            # Keeping positions on hold for a new fit is not really sensible
            if "pos" in self.workFit.fitter.params:
                self.workFit.fitter.SetParameter("pos", "free")
                hdtv.ui.msg("'pos' fit parameter reset to 'free'")
            self.workFit.spec = None

    def ActivateFit(self, ID, sid=None):
        """
        Copy markers of a already stored fit to workFit.

        Note: that a call to Store will overwrite the active fit with workFit
        """
        # for interactive use of this function
        if isinstance(ID, (str, int, float)):
            ID = hdtv.util.ID.ParseIds(ID, self)[0]
        if sid is None:
            sid = self.activeID
        if sid is None:
            hdtv.ui.warning("There is no active spectrum")
            return
        spec = self.dict[sid]
        spec.ActivateObject(ID)
        if ID is not None:
            # make a copy for workFit
            self.workFit = copy.copy(spec.GetActiveObject())
            self.workFit.active = True
            self.workFit.Draw(self.viewport)

    def StoreFit(self, ID=None):
        """
        Stores current workFit with ID.

        If no ID is given, the next free ID is used.
        The markers are kept in workFit, for possible re-use.
        """
        spec = self.workFit.spec
        # for interactive use of this function
        if isinstance(ID, (str, int, float)):
            ids = hdtv.util.ID.ParseIds(
                ID, self.dict[self.activeID], only_existent=False
            )
            if len(ids) > 1:
                raise hdtv.cmdline.HDTVCommandError("More than one ID given")
            ID = ids[0]
        if spec is None:
            hdtv.ui.warning("No fit available to store")
            return
        if ID is None:
            ID = spec.activeID
        ID = spec.Insert(self.workFit, ID)
        spec.dict[ID].active = False
        spec.ActivateObject(None)
        hdtv.ui.msg("Storing workFit with ID %s" % ID)
        self.workFit = copy.copy(self.workFit)
        self.workFit.active = True
        self.workFit.Draw(self.viewport)

    # Functions to handle workCut
    def ExecuteCut(self):
        spec = self.GetActiveObject()
        if spec is None:
            hdtv.ui.error("There is no active spectrum")
            return
        if not hasattr(spec, "matrix") or spec.matrix is None:
            hdtv.ui.error("Active spectrum does not belong to a matrix")
            return
        cutSpec = self.workCut.ExecuteCut(spec.matrix, spec.axis)
        if cutSpec is not None:
            ID = spec.matrix.ID
            self.Insert(cutSpec, ID=hdtv.util.ID(ID.major, 1010))

    def ClearCut(self):
        self.workCut.regionMarkers.Clear()
        self.workCut.bgMarkers.Clear()
        self.workCut.axis = None
        if self.workCut.spec is not None:
            self.Pop(self.Index(self.workCut.spec))

    def ActivateCut(self, ID):
        # for interactive use of this function
        if isinstance(ID, (str, int, float)):
            ID = hdtv.util.ID.ParseIds(ID, self)[0]
        spec = self.GetActiveObject()
        if spec is None:
            hdtv.ui.error("There is no active spectrum")
            return
        if not hasattr(spec, "matrix") or spec.matrix is None:
            hdtv.ui.error("Active spectrum does not belong to a matrix")
            return
        spec.matrix.ActivateObject(ID)
        if ID is not None:
            self.workCut = copy.copy(spec.matrix.GetActiveObject())
            self.workCut.active = True
            self.workCut.Draw(self.viewport)

    def StoreCut(self, ID=None):
        # for interactive use of this function
        if isinstance(ID, (str, int, float)):
            ID = hdtv.util.ID.ParseIds(ID, self)[0]
        spec = self.GetActiveObject()
        if spec is None:
            hdtv.ui.error("There is no active spectrum")
            return
        if not hasattr(spec, "matrix") or spec.matrix is None:
            hdtv.ui.error("Active spectrum does not belong to a matrix")
            return
        mat = spec.matrix
        ID = mat.Insert(self.workCut, ID)
        mat.dict[ID].active = False
        mat.ActivateObject(None)
        if self.workCut.spec is not None:
            spec = self.Pop(self.Index(self.workCut.spec))
            # give it a new color
            spec.color = hdtv.color.ColorForID(ID.major)
            self.Insert(spec, ID=hdtv.util.ID(mat.ID.major, ID.major))
        hdtv.ui.msg("Storing workCut with ID %s" % ID)
        self.workCut = copy.copy(self.workCut)
        self.workCut.active = True
        self.workCut.Draw(self.viewport)

    # Overwrite some functions of DrawableManager to do some extra work
    def ActivateObject(self, obj):
        """
        Activate Object and reset workFit when activating another spectrum
        """
        # for interactive use of this function
        if isinstance(obj, (str, int, float)):
            obj = hdtv.util.ID.ParseIds(obj, self)[0]
        if obj is not self.activeID:
            # reset workFit
            self.workFit.spec = None
            # deactivate a possible active fit
            spec = self.GetActiveObject()
            if spec is not None:
                spec.ActivateObject(None)
        super().ActivateObject(obj)

    def Pop(self, ID):
        """
        Pop spectrum with ID and activate prevID if ID was activeID
        """
        # for interactive use of this function
        if isinstance(ID, (str, int, float)):
            ID = hdtv.util.ID.ParseIds(ID, self)[0]
        if self.activeID is ID:
            self.ActivateObject(self.prevID)
        return super().Pop(ID)

    def ShowObjects(self, ids, clear=True):
        """
        Show spectra and make sure one of the visible objects is active
        """
        ids = super().ShowObjects(ids, clear)
        if self.activeID not in self.visible:
            if len(self.visible) > 0:
                self.ActivateObject(max(self.visible))
            else:
                self.ActivateObject(None)
        return ids

    def HideObjects(self, ids):
        """
        Hide spectra and make sure one of the visible objects is active
        """
        ids = super().HideObjects(ids)
        if self.activeID not in self.visible:
            if len(self.visible) > 0:
                self.ActivateObject(max(self.visible))
            else:
                self.ActivateObject(None)
        return ids

    def Clear(self):
        """
        Clear everything
        """
        self.workFit = Fit(Fitter(peakModel="theuerkauf", backgroundModel="polynomial"))
        self.workFit.active = True
        self.workFit.Draw(self.viewport)
        self.caldict = hdtv.cal.PositionCalibrationDict()
        return super().Clear()



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import ROOT
import hdtv.rootext.display

import hdtv.cal
import hdtv.color

from hdtv.drawable import Drawable
from hdtv.util import Position, LockViewport


class Marker(Drawable):
    """
    Markers in X or in Y direction

    Every marker contains a pair of positions as many markers come always in
    pairs to mark a region (like background markers). Of course it is also
    possible to have markers that consist of a single marker, then the second
    position is None.
    """

    def __init__(self, xytype, p1, color=hdtv.color.zoom, cal=None, connecttop=False):
        self._activeColor = color
        self._cal = cal
        self._dashed = False
        self.fixedInCal = True
        self.xytype = xytype
        self.connecttop = connecttop
        self.p1 = p1
        self.p2 = None
        Drawable.__init__(self, color=color, cal=cal)

    # p1 and p2 properties
    def _set_p(self, pos, p):
        if isinstance(pos, (float, int)):
            pos = hdtv.util.Position(pos, self.fixedInCal, self.cal)
        setattr(self, "_%s" % p, pos)

    def _get_p(self, p):
        return getattr(self, "_%s" % p)

    p1 = property(
        lambda self: self._get_p("p1"), lambda self, pos: self._set_p(pos, "p1")
    )
    p2 = property(
        lambda self: self._get_p("p2"), lambda self, pos: self._set_p(pos, "p2")
    )

    # color property
    def _set_color(self, color):
        # active color is given at creation and should not change
        self._passiveColor = hdtv.color.Highlight(color, active=False)
        if self.displayObj:
            if self.active:
                self.displayObj.SetColor(self._activeColor)
            else:
                self.displayObj.SetColor(self._passiveColor)

    def _get_color(self):
        return self._passiveColor

    color = property(_get_color, _set_color)

    # cal property
    def _set_cal(self, cal):
        self.p1.cal = cal
        if self.p2 is not None:
            self.p2.cal = cal
        Drawable._set_cal(self, cal)
        if self.displayObj:
            # call Refresh to carry the possible change
            # of p1.pos_uncal und p2.pos_uncal to the displayObj
            self.Refresh()

    def _get_cal(self):
        return self._cal

    cal = property(_get_cal, _set_cal)

    # dashed property
    def _set_dashed(self, state):
        self._dashed = state
        if self.displayObj:
            self.displayObj.SetDash(state, state)

    def _get_dashed(self):
        return self._dashed

    dashed = property(_get_dashed, _set_dashed)

    def Draw(self, viewport):
        """
        Draw the marker
        """
        if self.viewport and not self.viewport == viewport:
            # Marker can only be drawn to a single viewport
            raise RuntimeError("Marker cannot be realized on multiple viewports")
        self.viewport = viewport
        # adjust the position values for the creation of the makers
        # on the C++ side all values must be uncalibrated
        p1 = self.p1.pos_uncal
        if self.p2 is None:
            n = 1
            p2 = 0.0
        else:
            n = 2
            p2 = self.p2.pos_uncal
        # X or Y?
        if self.xytype == "X":
            constructor = ROOT.HDTV.Display.XMarker
        elif self.xytype == "Y":
            constructor = ROOT.HDTV.Display.YMarker
        if self.active:
            self.displayObj = constructor(n, p1, p2, self._activeColor)
        else:
            self.displayObj = constructor(n, p1, p2, self._passiveColor)
        # set some properties
        if self.ID is not None:
            self.displayObj.SetID(str(self.ID))
        self.displayObj.SetDash(self.dashed, self.dashed)
        if self.xytype == "X":
            # these properties make only sense on the X axis
            self.displayObj.SetConnectTop(self.connecttop)
            if self.cal:
                self.displayObj.SetCal(self.cal)
        self.displayObj.Draw(self.viewport)

    def Refresh(self):
        if not self.displayObj:
            return
        p1 = self.p1.pos_uncal
        if self.p2 is None:
            # on the C++ side all values must be uncalibrated
            self.displayObj.SetN(1)
            self.displayObj.SetPos(p1)
        else:
            # on the C++ side all values must be uncalibrated
            p2 = self.p2.pos_uncal
            self.displayObj.SetN(2)
            self.displayObj.SetPos(p1, p2)

    def __str__(self):
        if self.p2 is not None:
            return f"{self.xytype} marker at {self.p1} and {self.p2}"
        else:
            return f"{self.xytype} marker at {self.p1}"

    def FixInCal(self):
        self.fixedInCal = True
        self.p1.FixInCal()
        if self.p2 is not None:
            self.p2.FixInCal()

    def FixInUncal(self):
        self.fixedInCal = False
        self.p1.FixInUncal()
        if self.p2 is not None:
            self.p2.FixInUncal()


class MarkerCollection(list):
    """
    A collection of identical markers
    """

    def __init__(
        self, xytype, paired=False, maxnum=None, color=None, cal=None, connecttop=True
    ):
        list.__init__(self)
        self.viewport = None
        self.xytype = xytype
        self.paired = paired
        self.maxnum = maxnum
        self.connecttop = connecttop
        self.cal = cal
        self.ID = None
        # active color is defined at creation time of MarkerCollection
        self.activeColor = color
        # color = passiveColor can be changed
        self.color = hdtv.color.Highlight(color, active=False)
        self.active = True  # By default markers are active when newly created
        self.fixedInCal = True  # By default markers are fixed in calibrated space

    # delegate everything to the markers
    def __setattr__(self, name, value):
        for marker in self:
            if hasattr(marker, name):
                marker.__setattr__(name, value)
        self.__dict__[name] = value

    def Draw(self, viewport):
        self.viewport = viewport
        for marker in self:
            marker.Draw(self.viewport)

    def Show(self):
        for marker in self:
            marker.Show()

    def Hide(self):
        for marker in self:
            marker.Hide()

    def Refresh(self):
        for marker in self:
            marker.Refresh()

    def FixInCal(self):
        self.fixedInCal = True
        for marker in self:
            marker.FixInCal()

    def FixInUncal(self):
        self.fixedInCal = False
        for marker in self:
            marker.FixInUncal()

    def SetMarker(self, pos):
        """
        Set a marker to calibrated position pos, possibly completing a marker pair
        """
        if self.IsFull():
            pending = self.pop(0)
            pending.p1 = pos
            pending.p2 = None
            pending.Refresh()
            self.append(pending)
        elif self.IsPending():
            pending = self[-1]
            pending.p2 = pos
            pending.Refresh()
        else:
            m = Marker(self.xytype, pos, self.activeColor, self.cal, self.connecttop)
            m.color = self.color
            m.ID = self.ID
            m.active = self.active
            self.append(m)
            if self.viewport:
                m.Draw(self.viewport)

    def IsFull(self):
        """
        Checks if this MarkerCollection already contains the maximum number of
        markers allowed
        """
        if self.maxnum is None:
            return False
        if self.IsPending():
            return False
        return len(self) == self.maxnum

    def IsPending(self):
        """
        Checks if there is a single marker waiting to form a pair.
        Always returns False if this is a MarkerCollection with paired == False.
        """
        if not self.paired:
            return False
        return len(self) > 0 and self[-1].p2 is None

    def Clear(self):
        """
        Remove all markers from this collection
        """
        with LockViewport(self.viewport):
            while self:
                self.pop()

    def RemoveNearest(self, pos):
        """
        Remove the marker that is nearest to pos
        If one of the members of a marker pair is nearest to pos,
        both are removed
        """
        if len(self) == 0:
            hdtv.ui.warning("No marker available, no action taken")
            return
        index = dict()
        for m in self:
            p1 = m.p1.pos_cal
            diff = abs(pos - p1)
            index[diff] = m
            if self.paired and m.p2 is not None:
                p2 = m.p2.pos_cal
                diff = abs(pos - p2)
                index[diff] = m
        nearest = index[min(index.keys())]
        self.remove(nearest)
        self.Refresh()



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import sys
import os
import tempfile
import shutil

sys.path.append(os.path.join(os.path.dirname(__file__), "helpers"))


def pytest_addoption(parser):
    parser.addoption(
        "--force-rebuild", action="store_true", help="Force library rebuild."
    )


def pytest_configure(config):
    print("Update Root Include Path ...")
    import hdtv.rootext

    hdtv.rootext.UpdateRootIncludePath()
    os.environ["XDG_CACHE_HOME"] = tempfile.mkdtemp()
    if config.getoption("force_rebuild"):
        print("Force Library Rebuild ...")
        import hdtv.rootext.dlmgr

        hdtv.rootext.dlmgr.RebuildLibraries(hdtv.rootext.dlmgr.usrlibdir)


def pytest_sessionfinish(session, exitstatus):
    tmpdir = os.getenv("XDG_CACHE_HOME")
    if tmpdir != "" and os.path.exists(tmpdir):
        # fingers crossed
        shutil.rmtree(tmpdir)



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
This script contains some test cases for the writing and reading of fits to XML
if this test work fine, one also needs to test, how thing behave after changing
the calibration back and forth (see test_cal.py for that).
"""


import os

import pytest

from tests.helpers.utils import setup_io, redirect_stdout
from tests.helpers.fixtures import temp_file_compressed

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import __main__

import hdtv.session

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass
spectra = __main__.spectra

from hdtv.plugins.specInterface import spec_interface
from hdtv.plugins.fitInterface import fit_interface
from hdtv.plugins.fitlist import fitxml

testspectrum = os.path.join(os.path.curdir, "tests", "share", "osiris_bg.spc")


@pytest.fixture(autouse=True)
def prepare():
    fit_interface.ResetFitterParameters()
    hdtv.options.Set("table", "classic")
    hdtv.options.Set("uncertainties", "short")
    spec_interface.LoadSpectra(testspectrum)
    yield
    spectra.Clear()


def list_fit():
    f, ferr = setup_io(2)
    with redirect_stdout(f, ferr):
        fit_interface.ListFits()
        fit_interface.ListIntegrals()
    assert ferr.getvalue().strip() == ""
    return f.getvalue().strip()


def fit_write_and_save(filename):
    spectra.ExecuteFit()
    spectra.StoreFit()
    spectra.ClearFit()

    out_original = list_fit()

    print("Saving fits to file %s" % filename)
    fitxml.WriteXML(spectra.Get("0").ID, filename)
    print("Deleting all fits")
    spectra.Get("0").Clear()
    print("Reading fits from file %s" % filename)
    fitxml.ReadXML(spectra.Get("0").ID, filename)

    assert out_original == list_fit()


def test_fitxml_all_free_no_bg(temp_file_compressed):
    """
    all parameter free, just one peak, no background, theuerkauf model
    """
    fit_interface.SetPeakModel("theuerkauf")
    spectra.SetMarker("region", 1450)
    spectra.SetMarker("region", 1470)
    spectra.SetMarker("peak", 1460)
    fit_write_and_save(temp_file_compressed)


def test_fitxml_all_free_bg(temp_file_compressed):
    """
    all parameter free, just one peak, background
    """
    spectra.SetMarker("region", 500)
    spectra.SetMarker("region", 520)
    spectra.SetMarker("peak", 511)
    spectra.SetMarker("bg", 480)
    spectra.SetMarker("bg", 490)
    spectra.SetMarker("bg", 530)
    spectra.SetMarker("bg", 540)
    fit_write_and_save(temp_file_compressed)


def test_fitxml_all_free_multi_peak(temp_file_compressed):
    """
    all parameter free, more than one peak
    """
    spectra.SetMarker("region", 1395)
    spectra.SetMarker("region", 1415)
    spectra.SetMarker("peak", 1400)
    spectra.SetMarker("peak", 1410)
    spectra.SetMarker("bg", 1350)
    spectra.SetMarker("bg", 1355)
    spectra.SetMarker("bg", 1420)
    spectra.SetMarker("bg", 1425)
    fit_write_and_save(temp_file_compressed)


def test_fitxml_parameter_hold(temp_file_compressed):
    """
    one parameter status!=free, but equal for all peaks
    """
    spectra.SetMarker("region", 960)
    spectra.SetMarker("region", 975)
    spectra.SetMarker("peak", 965)
    spectra.SetMarker("peak", 970)
    spectra.SetMarker("bg", 950)
    spectra.SetMarker("bg", 955)
    spectra.SetMarker("bg", 980)
    spectra.SetMarker("bg", 985)
    fit_interface.SetFitterParameter("pos", "hold")
    fit_write_and_save(temp_file_compressed)


def test_fitxml_parameter_multi(temp_file_compressed):
    """
    different parameter status for each peak
    """
    spectra.SetMarker("region", 1750)
    spectra.SetMarker("region", 1780)
    spectra.SetMarker("peak", 1765)
    spectra.SetMarker("peak", 1770)
    spectra.SetMarker("bg", 1700)
    spectra.SetMarker("bg", 1710)
    spectra.SetMarker("bg", 1800)
    spectra.SetMarker("bg", 1810)
    fit_interface.SetFitterParameter("pos", "hold,free")
    fit_write_and_save(temp_file_compressed)


def test_fitxml_eepeak(temp_file_compressed):
    """
    ee peak (just proof of concept, not a thorough test)
    """
    fit_interface.SetPeakModel("ee")
    spectra.SetMarker("region", 1115)
    spectra.SetMarker("region", 1125)
    spectra.SetMarker("peak", 1120)
    fit_write_and_save(temp_file_compressed)



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import copy
from uncertainties import ufloat

import ROOT
import hdtv.color
import hdtv.cal
import hdtv.integral
import hdtv.ui

from hdtv.drawable import Drawable
from hdtv.marker import MarkerCollection
from hdtv.util import Pairs, ID, Table, LockViewport
from hdtv.weakref_proxy import weakref


class Fit(Drawable):
    """
    Fit object

    This Fit object is the graphical representation of a fit in HDTV.
    It contains the marker lists and the functions. The actual interface to
    the C++ fitting routines is the class Fitter.

    All internal values (fit parameters, fit region, peak list) are in
    uncalibrated units.
    """

    # List of hook functions to be called before/after FitPeakFunc()
    # These hook functions should accept a reference to the Fit class
    # that calls them as first argument
    FitPeakPreHooks = list()
    FitPeakPostHooks = list()
    showDecomp = False  # Default value for decomposition status

    def __init__(self, fitter, color=None, cal=None):
        self.regionMarkers = MarkerCollection(
            "X", paired=True, maxnum=1, color=hdtv.color.region, cal=cal
        )
        self.peakMarkers = MarkerCollection(
            "X", paired=False, maxnum=None, color=hdtv.color.peak, cal=cal
        )
        self.bgMarkers = MarkerCollection(
            "X", paired=True, maxnum=None, color=hdtv.color.bg, cal=cal
        )
        self.fitter = fitter
        self.peaks = []
        self.chi = None
        self.bgChi = None
        self.bgParams = []
        self._showDecomp = Fit.showDecomp
        self.dispPeakFunc = None
        self.dispBgFunc = None
        Drawable.__init__(self, color, cal)
        self._spec = None
        self.active = False
        self.integral = None

    # ID property
    def _get_ID(self):
        return self._ID

    def _set_ID(self, ID):
        self._ID = ID
        if ID is not None:
            self.peakMarkers.ID = ID
        else:
            self.peakMarkers.ID = ID

    ID = property(_get_ID, _set_ID)

    # cal property
    def _set_cal(self, cal):
        self._cal = hdtv.cal.MakeCalibration(cal)
        with LockViewport(self.viewport):
            self.peakMarkers.cal = self._cal
            self.regionMarkers.cal = self._cal
            self.bgMarkers.cal = self._cal
            if self.dispPeakFunc:
                self.dispPeakFunc.SetCal(self._cal)
            if self.dispBgFunc:
                self.dispBgFunc.SetCal(self._cal)
            for peak in self.peaks:
                peak.cal = self._cal

    def _get_cal(self):
        return self._cal

    cal = property(_get_cal, _set_cal)

    # color property
    def _set_color(self, color):
        # we only need the passive color for fits
        self._passiveColor = hdtv.color.Highlight(color, active=False)
        with LockViewport(self.viewport):
            self.peakMarkers.color = color
            self.regionMarkers.color = color
            self.bgMarkers.color = color
            for peak in self.peaks:
                peak.color = color
            self.Show()

    def _get_color(self):
        return self._passiveColor

    color = property(_get_color, _set_color)

    # active property
    def _set_active(self, state):
        with LockViewport(self.viewport):
            self._active = state
            self.peakMarkers.active = state
            self.regionMarkers.active = state
            self.bgMarkers.active = state
            for peak in self.peaks:
                peak.active = state
            self.Show()

    def _get_active(self):
        return self._active

    active = property(_get_active, _set_active)

    # spec property
    def _set_spec(self, spec):
        if hasattr(self, "spec") and self._spec == spec:
            return
        # use weakref to prevent problems with cyclic reference
        self._spec = weakref(spec)
        self.Erase()
        if spec is None:
            self.FixMarkerInCal()
            self.cal = None
            self.color = None
        else:
            self.color = spec.color
            self.cal = spec.cal
            self.FixMarkerInUncal()

    def _get_spec(self):
        return self._spec

    spec = property(_get_spec, _set_spec)

    # ids property to get the ids of the peaks
    @property
    def ids(self):
        ids = list()
        for i in range(len(self.peaks)):
            ids.append(ID(self.ID.major, i))
        return ids

    def __str__(self):
        """
        show fit results in a nice table
        """
        (objects, params) = self.ExtractParams()
        header = (
            "WorkFit on spectrum: " + str(self.spec.ID) + " (" + self.spec.name + ")"
        )
        footer = "\n" + str(len(objects)) + " peaks in WorkFit"
        table = Table(
            objects, list(params), sortBy="id", extra_header=header, extra_footer=footer
        )
        return str(table)

    def print_integral(self):
        """
        show integral of fit regions in a nice table
        """
        (objects, params) = self.ExtractIntegralParams("all")
        header = (
            "Integrals of WorkFit regions on spectrum: "
            + str(self.spec.ID)
            + " ("
            + self.spec.name
            + ")"
        )
        table = Table(objects, list(params), sortBy="id", extra_header=header)
        return str(table)

    def formatted_str(self, verbose=True):
        """
        show fit results in a more tv like way

        TODO: Since we do not use this any longer, we may as well remove it.
        """
        i = 0
        text = ""
        for peak in self.peaks:
            text += ("\nPeak %d:" % i).ljust(10)
            peakstr = peak.formatted_str(verbose).strip()
            peakstr = peakstr.split("\n")
            peakstr = ("\n".ljust(10)).join(peakstr)
            text += peakstr
            i += 1
        text += "\n\n chi² of fit: %d" % self.chi
        return text

    def ExtractParams(self):
        """
        Helper function for use for printing fit results in a nice table

        Return values:
            peaklist: a list of dicts for each peak in the fit
            params  : a ordered list of valid parameter names
        """
        peaklist = list()
        params = ["id", "stat", "chi"]
        # Get peaks
        for peak in self.peaks:
            thispeak = dict()
            thispeak["chi"] = "%d" % self.chi
            if self.ID is None:
                thispeak["id"] = hdtv.util.ID(None, self.peaks.index(peak))
            else:
                thispeak["id"] = hdtv.util.ID(self.ID.major, self.peaks.index(peak))
            thispeak["stat"] = ""
            if self.active:
                thispeak["stat"] += "A"
            if self.ID in self.spec.visible or self.ID is None:  # ID of workFit is None
                thispeak["stat"] += "V"
            # get parameter of this fit
            for p in self.fitter.peakModel.fValidParStatus.keys():
                if p == "pos":
                    # Store channel additionally to position
                    thispeak["channel"] = getattr(peak, "pos")
                    if "channel" not in params:
                        params.append("channel")
                # Use calibrated values of params if available
                p_cal = p + "_cal"
                if hasattr(peak, p_cal):
                    thispeak[p] = getattr(peak, p_cal)
                if p not in params:
                    params.append(p)
            # add extra params
            for p in list(peak.extras.keys()):
                thispeak[p] = peak.extras[p]
                if p not in params:
                    params.append(p)
            # Calculate normalized volume if efficiency calibration is present
            # FIXME: does this belong here?
            # if self.spec.effCal is not None:
            #    volume = thispeak["vol"]
            #    energy = thispeak["pos"]
            #    norm_volume = volume / self.spec.effCal(energy)
            #    par = "vol/eff"
            #    par_index = params.index("vol") + 1
            #    params.insert(par_index, par)
            #    thispeak[par] = norm_volume
            peaklist.append(thispeak)
        return (peaklist, params)

    def ExtractIntegralParams(self, integral_type="auto"):
        """
        Helper function for use for printing fit results in a nice table

        Return values:
            integrallist : a list of dicts for each peak in the fit
            params       : a ordered list of valid parameter names
        """
        integrallist = list()
        params = ["id", "stat", "type"]

        # if not self.integral:
        #    region = [self.regionMarkers[0].p1.pos_uncal,
        #              self.regionMarkers[0].p2.pos_uncal]
        #    self.integral = hdtv.integral.Integrate(
        #        self.spec, self.fitter.bgFitter, region)
        integrals = self.integral

        if integral_type == "all":
            integral_types = ["tot", "bg", "sub"]
        elif integral_type == "auto":
            if integrals["sub"]:
                integral_types = ["sub"]
            else:
                integral_types = ["tot"]
        else:
            integral_types = [integral_type]

        for int_type, integral in integrals.items():
            if not integral or not int_type in integral_types:
                continue
            int_res = dict(integral["uncal"])

            int_res["type"] = int_type

            if self.ID is None:
                int_res["id"] = hdtv.util.ID(None, None)
            else:
                int_res["id"] = hdtv.util.ID(self.ID.major, None)
            int_res["stat"] = ""

            if self.active:
                int_res["stat"] += "A"
            if self.ID in self.spec.visible or self.ID is None:  # ID of workFit is None
                int_res["stat"] += "V"

            for p in integral["uncal"].keys():
                if p not in params:
                    params.append(p)

            if self.spec.cal:
                # rename pos to channel in uncal
                int_res["channel"] = int_res.pop("pos")
                integral.pop("cal", None)
                # Make sure that calibration is up to date
                integral["cal"] = hdtv.integral.calibrate_integral(
                    integral, self.spec.cal
                )["cal"]
                for key, value in integral["cal"].items():
                    if key == "vol":
                        continue
                    p = key if (key == "pos") else key + "_cal"
                    int_res[p] = value
                    if p not in params:
                        params.append(p)

            integrallist.append(int_res)
        return (integrallist, params)

    def ChangeMarker(self, mtype, pos, action):
        """
        Set a new marker or remove a marker.

        action : "put" or "remove"
        mtype  : "bg", "region", "peak"
        """
        self.spec = None
        markers = getattr(self, "%sMarkers" % mtype)
        if action == "set":
            markers.SetMarker(pos)
        if action == "remove":
            markers.RemoveNearest(pos)

    def FixMarkerInCal(self):
        """
        Fix marker in calibrated space
        """
        for m in [self.bgMarkers, self.regionMarkers, self.peakMarkers]:
            m.FixInCal()

    def FixMarkerInUncal(self):
        """
        Fix marker in uncalibrated space
        """
        for m in [self.bgMarkers, self.regionMarkers, self.peakMarkers]:
            m.FixInUncal()

    def _get_background_pairs(self):
        if self.bgMarkers.IsPending():
            hdtv.ui.warning("Not all background regions are closed.")
        backgrounds = Pairs()
        for m in self.bgMarkers:
            try:
                backgrounds.add(m.p1.pos_uncal, m.p2.pos_uncal)
            except AttributeError:
                pos = m.p1.pos_cal if m.p1.pos_cal else m.p2.pos_cal
                hdtv.ui.warning(
                    "Background region at {:.2f} without second marker was ignored".format(
                        pos
                    )
                )
        return backgrounds

    def FitBgFunc(self, spec=None):
        """
        Do the background fit and extract the function for display
        Note: You still need to call Draw afterwards.
        """
        if spec is not None:
            self.spec = spec
        self.Erase()
        hdtv.ui.debug("Fitting background")
        # fit background
        if len(self.bgMarkers) > 0:
            backgrounds = self._get_background_pairs()

            try:
                self.fitter.FitBackground(spec=self.spec, backgrounds=backgrounds)
                func = self.fitter.bgFitter.GetFunc()
                self.dispBgFunc = ROOT.HDTV.Display.DisplayFunc(func, hdtv.color.bg)
                self.dispBgFunc.SetCal(self.cal)
                self.bgChi = self.fitter.bgFitter.GetChisquare()
                self.bgParams = []
                nparams = self.fitter.bgFitter.GetNparams()
                for i in range(0, nparams):
                    self.bgParams.append(
                        ufloat(
                            self.fitter.bgFitter.GetCoeff(i),
                            self.fitter.bgFitter.GetCoeffError(i),
                        )
                    )
            except ValueError:
                raise hdtv.cmdline.HDTVCommandAbort("Background fit failed.")

    def FitPeakFunc(self, spec=None):
        """
        Do the actual peak fit and extract the functions for display
        Note: You still need to call Draw afterwards.
        """
        # Call pre hooks
        for func in Fit.FitPeakPreHooks:
            func(self)

        if spec is not None:
            self.spec = spec
        self.Erase()
        # fit background
        if len(self.bgMarkers) > 0:
            backgrounds = self._get_background_pairs()
            try:
                self.fitter.FitBackground(spec=self.spec, backgrounds=backgrounds)
            except ValueError:
                raise hdtv.cmdline.HDTVCommandAbort("Background fit failed.")
        # fit peaks
        if len(self.peakMarkers) > 0 and self.regionMarkers.IsFull():
            region = sorted(
                [self.regionMarkers[0].p1.pos_uncal, self.regionMarkers[0].p2.pos_uncal]
            )
            # remove peak marker that are outside of region
            for m in self.peakMarkers[:]:
                # we need to loop over a copy here,
                # otherwise we get out of sync after deleting items
                if m.p1.pos_uncal < region[0] or m.p1.pos_uncal > region[1]:
                    self.peakMarkers.remove(m)
            peaks = sorted(m.p1.pos_uncal for m in self.peakMarkers)
            self.fitter.FitPeaks(spec=self.spec, region=region, peaklist=peaks)
            # get background function
            self.bgParams = []
            nparams = self.fitter.backgroundModel.fParStatus["nparams"]
            if not self.fitter.bgFitter:
                for i in range(nparams):
                    self.bgParams.append(
                        ufloat(
                            self.fitter.peakFitter.GetIntBgCoeff(i),
                            self.fitter.peakFitter.GetIntBgCoeffError(i),
                        )
                    )
            else:
                # external background
                self.bgChi = self.fitter.bgFitter.GetChisquare()
                for i in range(nparams):
                    self.bgParams.append(
                        ufloat(
                            self.fitter.bgFitter.GetCoeff(i),
                            self.fitter.bgFitter.GetCoeffError(i),
                        )
                    )
            func = self.fitter.peakFitter.GetBgFunc()
            self.dispBgFunc = ROOT.HDTV.Display.DisplayFunc(func, hdtv.color.bg)
            self.dispBgFunc.SetCal(self.cal)
            # get peak function
            func = self.fitter.peakFitter.GetSumFunc()
            self.dispPeakFunc = ROOT.HDTV.Display.DisplayFunc(func, hdtv.color.region)
            self.dispPeakFunc.SetCal(self.cal)
            self.chi = self.fitter.peakFitter.GetChisquare()
            # create peak list
            for i in range(0, self.fitter.peakFitter.GetNumPeaks()):
                cpeak = self.fitter.peakFitter.GetPeak(i)
                peak = self.fitter.peakModel.CopyPeak(cpeak, hdtv.color.peak, self.cal)
                self.peaks.append(peak)
            # in some rare cases it can happen that peaks change position
            # while doing the fit, thus we have to sort here
            self.peaks.sort()
            # update peak markers
            for (marker, peak) in zip(self.peakMarkers, self.peaks):
                # Marker is fixed in uncalibrated space
                marker.p1.pos_uncal = peak.pos.nominal_value

        # Call post hooks
        for func in Fit.FitPeakPostHooks:
            func(self)

    def Restore(self, spec):
        # do not call Erase() while setting spec!
        self._spec = weakref(spec)
        self.cal = spec.cal
        self.color = spec.color
        self.FixMarkerInUncal()
        if len(self.bgMarkers) > 0 and not self.bgMarkers.IsPending():
            backgrounds = Pairs()
            for m in self.bgMarkers:
                backgrounds.add(m.p1.pos_uncal, m.p2.pos_uncal)
            self.fitter.RestoreBackground(
                backgrounds=backgrounds, params=self.bgParams, chisquare=self.bgChi
            )
        region = sorted(
            [self.regionMarkers[0].p1.pos_uncal, self.regionMarkers[0].p2.pos_uncal]
        )
        if self.peaks:
            self.fitter.RestorePeaks(
                cal=self.cal,
                region=region,
                peaks=self.peaks,
                chisquare=self.chi,
                coeffs=self.bgParams,
            )
            # get background function
            func = self.fitter.peakFitter.GetBgFunc()
            self.dispBgFunc = ROOT.HDTV.Display.DisplayFunc(func, self.color)
            self.dispBgFunc.SetCal(self.cal)
            # get peak function
            func = self.fitter.peakFitter.GetSumFunc()
            self.dispPeakFunc = ROOT.HDTV.Display.DisplayFunc(func, self.color)
            self.dispPeakFunc.SetCal(self.cal)

            # restore display functions of single peaks
            for i in range(0, self.fitter.peakFitter.GetNumPeaks()):
                cpeak = self.fitter.peakFitter.GetPeak(i)
                func = cpeak.GetPeakFunc()
                self.peaks[i].displayObj = ROOT.HDTV.Display.DisplayFunc(
                    func, self.color
                )
                self.peaks[i].displayObj.SetCal(self.cal)

            # create non-existant integral
            if not self.integral:
                self.integral = hdtv.integral.Integrate(
                    self.spec, self.fitter.bgFitter, region
                )

    def Draw(self, viewport):
        """
        Draw
        """
        if self.viewport and not self.viewport == viewport:
            # Unlike the Display object of the underlying implementation,
            # python objects can only be drawn on a single viewport
            raise RuntimeError("Object can only be drawn on a single viewport")
        self.viewport = viewport
        with LockViewport(self.viewport):
            # draw the markers (do this after the fit,
            # because the fit updates the position of the peak markers)
            self.peakMarkers.Draw(self.viewport)
            self.regionMarkers.Draw(self.viewport)
            self.bgMarkers.Draw(self.viewport)
            # draw fit func, if available
            if self.dispPeakFunc:
                self.dispPeakFunc.Draw(self.viewport)
            if self.dispBgFunc:
                self.dispBgFunc.Draw(self.viewport)
            # draw peaks
            for peak in self.peaks:
                peak.color = self.color
                peak.Draw(self.viewport)
            self.Show()

    def Refresh(self):
        """
        Refresh
        """
        if self.spec is None:
            return
        # repeat the fits
        if self.dispPeakFunc:
            # this includes the background fit
            self.FitPeakFunc(self.spec)
        elif self.dispBgFunc:
            # maybe there was only a background fit
            self.FitBgFunc(self.spec)
        if not self.viewport:
            return
        with LockViewport(self.viewport):
            # draw fit func, if available
            if self.dispPeakFunc:
                self.dispPeakFunc.Draw(self.viewport)
            if self.dispBgFunc:
                self.dispBgFunc.Draw(self.viewport)
            # draw peaks
            for peak in self.peaks:
                peak.Draw(self.viewport)
            # draw the markers (do this after the fit,
            # because the fit updates the position of the peak markers)
            self.peakMarkers.Refresh()
            self.regionMarkers.Refresh()
            self.bgMarkers.Refresh()
            self.Show()

    def Erase(self, bg_only=False):
        """
        Erase previous fit. NOTE: the fitter is *not* resetted
        """
        # remove bg fit
        self.dispBgFunc = None
        self.fitter.bgFitter = None
        self.bgParams = []
        self.bgChi = None
        if not bg_only:
            # remove peak fit
            self.dispPeakFunc = None
            self.peaks = []
            self.chi = None

    def ShowAsWorkFit(self):
        if not self.viewport:
            return
        with LockViewport(self.viewport):
            # all markers in active state and solid
            for mc in [self.regionMarkers, self.peakMarkers, self.bgMarkers]:
                mc.active = True
                mc.dashed = False
                mc.Show()
            # coloring
            if self.dispPeakFunc:
                self.dispPeakFunc.SetColor(hdtv.color.region)
                self.dispPeakFunc.Show()
            if self.dispBgFunc:
                self.dispBgFunc.SetColor(hdtv.color.bg)
                self.dispBgFunc.Show()
            # peak list
            for peak in self.peaks:
                peak.color = hdtv.color.peak
                if self._showDecomp:
                    peak.Show()
                else:
                    peak.Hide()

    def ShowAsPending(self):
        if not self.viewport:
            return
        with LockViewport(self.viewport):
            # all markers in passive state and dashed
            for mc in [self.regionMarkers, self.peakMarkers, self.bgMarkers]:
                mc.active = False
                mc.dashed = True
                mc.Show()
            # coloring
            if self.dispPeakFunc:
                self.dispPeakFunc.SetColor(self.color)
                self.dispPeakFunc.Show()
            if self.dispBgFunc:
                self.dispBgFunc.SetColor(self.color)
                self.dispBgFunc.Show()
            # peak list
            for peak in self.peaks:
                peak.color = self.color
                if self._showDecomp:
                    peak.Show()
                else:
                    peak.Hide()

    def ShowAsPassive(self):
        if not self.viewport:
            return
        with LockViewport(self.viewport):
            # marker in passive State and not dashed
            for mc in [self.regionMarkers, self.peakMarkers, self.bgMarkers]:
                mc.active = False
                mc.dashed = False
            # only show peakMarkers
            self.regionMarkers.Hide()
            self.peakMarkers.Show()
            self.bgMarkers.Hide()
            # coloring
            if self.dispPeakFunc:
                self.dispPeakFunc.SetColor(self.color)
                self.dispPeakFunc.Show()
            if self.dispBgFunc:
                self.dispBgFunc.SetColor(self.color)
                self.dispBgFunc.Show()
            # peak list
            for peak in self.peaks:
                peak.color = self.color
                if self._showDecomp:
                    peak.Show()
                else:
                    peak.Hide()

    def Show(self):
        if not self.viewport:
            return
        if self.active:
            if self.ID is None:
                self.ShowAsWorkFit()
            else:
                self.ShowAsPending()
        else:
            self.ShowAsPassive()

    def Hide(self):
        if not self.viewport:
            return
        with LockViewport(self.viewport):
            self.peakMarkers.Hide()
            self.regionMarkers.Hide()
            self.bgMarkers.Hide()
            if self.dispPeakFunc:
                self.dispPeakFunc.Hide()
            if self.dispBgFunc:
                self.dispBgFunc.Hide()
            for peak in self.peaks:
                peak.Hide()

    def __copy__(self):
        """
        Create new fit with identical markers
        """
        new = Fit(copy.copy(self.fitter), cal=self.cal, color=self.color)
        new.FixMarkerInCal()
        for marker in self.bgMarkers:
            new.bgMarkers.SetMarker(marker.p1.pos_cal)
            if marker.p2:
                new.bgMarkers.SetMarker(marker.p2.pos_cal)
        for marker in self.regionMarkers:
            new.regionMarkers.SetMarker(marker.p1.pos_cal)
            if marker.p2:
                new.regionMarkers.SetMarker(marker.p2.pos_cal)
        for marker in self.peakMarkers:
            new.peakMarkers.SetMarker(marker.p1.pos_cal)
            if marker.p2:
                new.peakMarkers.SetMarker(marker.p2.pos_cal)
        return new

    def __eq__(self, other):
        return self.xdimensions == other.xdimensions

    def __ne__(self, other):
        return self.xdimensions != other.xdimensions

    def __gt__(self, other):
        return self.xdimensions > other.xdimensions

    def __lt__(self, other):
        return self.xdimensions < other.xdimensions

    def __ge__(self, other):
        return self.xdimensions >= other.xdimensions

    def __le__(self, other):
        return self.xdimensions <= other.xdimensions

    @property
    def xdimensions(self):
        """
        Return dimensions of fit in x-direction

        returns: tuple (x_start, x_end)
        """
        markers = list()
        # Get maximum of region markers, peak markers and peaks
        for r in self.regionMarkers:
            markers.append(r.p1.pos_cal)
            markers.append(r.p2.pos_cal)
        for p in self.peakMarkers:
            markers.append(p.p1.pos_cal)
        for p in self.peaks:
            markers.append(p.pos_cal)
        for b in self.bgMarkers:
            markers.append(b.p1.pos_cal)
            markers.append(b.p2.pos_cal)
        # calulate region limits
        if not markers:  # No markers
            return None
        else:
            x_start = min(markers)
            x_end = max(markers)
            return (x_start, x_end)

    def SetDecomp(self, stat=True):
        """
        Sets whether to display a decomposition of the fit
        """
        self._showDecomp = stat
        if stat:
            for peak in self.peaks:
                peak.Show()
        else:
            for peak in self.peaks:
                peak.Hide()



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import re

import pytest

from tests.helpers.utils import redirect_stdout, hdtvcmd

import hdtv.cmdline
import hdtv.options
import hdtv.plugins.config

# populate with a few options
import hdtv.plugins.dblookup


@pytest.fixture(autouse=True)
def prepare():
    for variable in [
        "table",
        "database.db",
        "database.auto_lookup",
        "database.fuzziness",
        "database.sort_key",
        "database.sort_reverse",
    ]:
        hdtv.options.Reset(variable)
    yield


def test_cmd_config_show():
    f, ferr = hdtvcmd("config show")
    out = f.split("\n")
    assert len(out) > 2
    for line in out:
        assert ": " in line
    assert ferr == ""


@pytest.mark.parametrize(
    "variable, value",
    [
        ("table", "modern"),
        ("database.db", "PGAAlib_IKI2000"),
        ("database.auto_lookup", "False"),
        ("database.fuzziness", "1.0"),
        ("database.sort_key", "None"),
        ("database.sort_reverse", "False"),
    ],
)
def test_cmd_config_show_variable(variable, value):
    f, ferr = hdtvcmd("config show " + variable)
    out = f.split("\n")
    assert len(out) == 1
    res_variable, res_value = re.search("<b>(.*)</b>: (.*)", out[0]).groups()
    assert variable in res_variable
    assert res_value == value
    assert ferr == ""


@pytest.mark.parametrize(
    "variable, value",
    [
        ("database.db", "promptgammas"),
        ("database.auto_lookup", "True"),
        ("database.fuzziness", "5.0"),
        ("database.sort_key", "energy"),
        ("database.sort_reverse", "True"),
    ],
)
def test_cmd_config_set(variable, value):
    hdtvcmd(f"config set {variable} {value}")
    assert str(hdtv.options.Get(variable)) == value


@pytest.mark.parametrize(
    "variable, value, default",
    [
        ("database.db", "promptgammas", "PGAAlib_IKI2000"),
        ("database.auto_lookup", "True", "False"),
        ("database.fuzziness", "5.0", "1.0"),
        ("database.sort_key", "energy", "None"),
        ("database.sort_reverse", "True", "False"),
    ],
)
def test_cmd_config_reset(variable, value, default):
    hdtv.options.Set(variable, value)
    hdtvcmd(f"config reset {variable}")
    assert str(hdtv.options.Get(variable)) == default


@pytest.mark.parametrize(
    "bool_str, bool_val",
    [
        ("true", True),
        ("True", True),
        ("tRuE", True),
        ("false", False),
        ("False", False),
        ("fAlSe", False),
    ],
)
def test_parse_bool(bool_str, bool_val):
    assert hdtv.options.parse_bool(bool_str) == bool_val


def test_register_option():
    opt = hdtv.options.Option(default="default")
    hdtv.options.RegisterOption("test.option", opt)
    assert hdtv.options.Get("test.option") == "default"


def test_register_option_fail():
    error_occurred = False
    opt = hdtv.options.Option(default="default")
    hdtv.options.RegisterOption("test.fail", opt)
    try:
        hdtv.options.RegisterOption("test.fail", opt)
    except RuntimeError:
        error_occurred = True
    assert error_occurred
    assert hdtv.options.Get("test.fail") == "default"



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2019  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import re
import os
import shlex
from itertools import count
import contextlib
from html import escape
from html.parser import HTMLParser
from typing import Optional

import numpy as np
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.shortcuts import PromptSession
from prompt_toolkit.keys import Keys

import hdtv.options
import hdtv.ui
import hdtv.dummy


def Indent(s, indent=" "):
    """
    Re-format a (possibly multi-line) string such that each line is indented.
    """
    return indent + ("\n" + indent).join(s.splitlines()) + "\n"


def GetCompleteOptions(begin, options):
    l = len(begin)
    return [o + " " for o in options if o[0:l] == begin]


class MLStripper(HTMLParser):
    def __init__(self):
        super().__init__()
        self.reset()
        self.fed = []

    def handle_data(self, d):
        self.fed.append(d)

    def get_data(self):
        return "".join(self.fed)


def strip_tags(html):
    s = MLStripper()
    s.feed(html)
    return s.get_data()


class TxtFile:
    """
    Handle txt files, ignoring commented lines
    """

    def __init__(self, filename, mode="r"):

        self.lines = list()
        self.linos = list()
        self.mode = mode
        # TODO: this has to be handled properly (escaped whitespaces, etc...)
        filename = str(filename).rstrip()
        self.filename = os.path.expanduser(filename)
        self.fd = None

    def read(self, verbose=False):
        """
        Read text file to self.lines

        Comments are stripped and lines seperated by '\' are automatically
        concatenated
        """
        try:
            self.fd = open(self.filename, self.mode)
            prev_line = ""
            number = 0
            for line in self.fd:
                number += 1
                line = line.rstrip("\r\n ")
                # line is continued on next line
                if len(line) > 0 and line[-1] == "\\":
                    prev_line += line.rstrip("\\")
                    continue
                else:
                    if prev_line != "":
                        line = prev_line + " " + line
                    prev_line = ""
                if verbose:
                    hdtv.ui.msg("file> " + str(line))

                # Strip comments
                line = remove_comments(line)
                if line.strip() == "":
                    continue
                self.lines.append(line)
                self.linos.append(number)

        except OSError as msg:
            raise OSError("Error opening file:" + str(msg))
        # except BaseException:  # Let MainLoop handle other exceptions
        #    raise
        finally:
            if self.fd is not None:
                self.fd.close()

    def write(self):
        """
        Write lines stored in self.lines to text file

        Newlines are automatically appended if necessary
        """
        try:
            self.fd = open(self.filename, self.mode)
            for line in self.lines:
                line.rstrip("\r\n ")
                line += os.linesep
                self.fd.write(line)
        except OSError as msg:
            raise OSError("Error opening file: %s" % msg)
        except BaseException:
            raise
        finally:
            if self.fd is not None:
                self.fd.close()


class Pairs(list):
    """
    List of pair values

    conv_func: conversion function to be called before storage of pair
    """

    # default conversion is "identity" -> No conversion
    def __init__(self, conv_func=lambda x: x):

        super().__init__()
        self.conv_func = conv_func  # Conversion function, e.g. float

    def add(self, x, y):
        """
        Add a pair
        """
        self.append([self.conv_func(x), self.conv_func(y)])

    def remove(self, pair):
        """
        TODO
        """
        pass

    def fromFile(self, fname, sep=None):
        """
        Read pairs from file
        """
        file = TxtFile(fname)
        file.read()
        for line in file.lines:
            pair = line.split(sep)
            try:
                self.add(pair[0], pair[1])
            except (ValueError, IndexError):
                hdtv.ui.error("Invalid Line in", fname, ":", line)

    def fromLists(self, list1, list2):
        """
        Create pairs from to lists by assigning corresponding indices
        """
        if len(list1) != len(list2):
            hdtv.ui.error("Lists for Pairs.fromLists() are of different length")
            return False

        for item1, item2 in zip(list1, list2):
            self.add(item1, item2)


opt_table = hdtv.options.Option(
    default="modern",
    parse=hdtv.options.parse_choices(["classic", "simple", "grid", "modern"]),
)
hdtv.options.RegisterOption("table", opt_table)

opt_uncertainties = hdtv.options.Option(
    default="short", parse=hdtv.options.parse_choices(["short", "pretty", "long"])
)
hdtv.options.RegisterOption("uncertainties", opt_uncertainties)


class Table:
    """
    Class to store tables

    data: iterable that contains 'keys' as attributes or dict entries
    """

    def __init__(
        self,
        data,
        keys,
        header=None,
        ignoreEmptyCols=True,
        sortBy=None,
        reverseSort=False,
        raw_columns=None,
        extra_header=None,
        extra_footer=None,
    ):
        self.extra_header = extra_header
        self.extra_footer = extra_footer
        self.sortBy = sortBy
        self.keys = keys
        self.data = list()

        self.style = style = hdtv.options.Get("table")

        if style == "classic":
            self.col_sep_char = "|"
            self.empty_field = "-"
            self.header_sep_char = "-"
            self.header_sep_char_sorted = "~"
            self.crossing_char = "+"
        elif style == "simple":
            self.col_sep_char = "\t"
            self.empty_field = ""
            self.header_sep_char = ""
            self.header_sep_char_sorted = ""
            self.crossing_char = ""
        elif style == "grid":
            self.col_sep_char = "│"
            self.empty_field = ""
            self.header_sep_char = "─"
            self.header_sep_char_sorted = "╌"
            self.crossing_char = "┼"
        else:  # default style 'modern'
            self.col_sep_char = ""
            self.empty_field = ""
            self.header_sep_char = "─"
            self.header_sep_char_sorted = "━"
            self.crossing_char = ""

        self._width = 0  # Width of table
        self._col_width = list()  # width of columns

        self.raw_columns = raw_columns

        # One additional "border column"
        self._ignore_col = [ignoreEmptyCols for i in range(0, len(keys) + 1)]

        self.read_data(data, keys, header)
        if sortBy is not None:
            self.sort_data(sortBy, reverseSort)

    @property
    def num_columns(self):
        return len(self.keys)

    @property
    def num_rows(self):
        return len(self.data)

    def read_data(self, data, keys, header=None):
        self.data = list()
        for d in data:
            tmp = dict()
            for k in keys:
                try:
                    tmp[k] = d[k]
                except KeyError:
                    # there is a missing value in the data
                    tmp[k] = None
                except TypeError:
                    # Data is no dict so we try to read attribute
                    tmp[k] = getattr(d, k)
            self.data.append(tmp)
        # header
        if header is None:
            self.header = keys
        else:
            self.header = header
        # initial width of columns
        for header in self.header:
            self._col_width.append(len(str(header)) + 2)

    def build_lines(self):
        lines = list()
        for d in self.data:
            line = list()
            for i, key in enumerate(self.keys):
                try:
                    value = d[key]
                    # deal with ufloats
                    try:
                        if value.value is None:
                            value = ""
                    except BaseException:
                        pass
                    if value is None:
                        value = ""
                    else:
                        try:
                            if hdtv.options.Get("uncertainties") == "short":
                                value = f"{value:S}"
                            elif hdtv.options.Get("uncertainties") == "pretty":
                                value = f"{value:P}"
                            else:
                                value = f"{value:.4u}".replace("+/-", " ")
                        except BaseException:
                            value = str(value)
                    if value != "":  # We have values in this columns -> don't ignore it
                        self._ignore_col[i] = False

                    line.append(value)
                    if self.raw_columns and key in self.raw_columns:
                        self._col_width[i] = max(
                            self._col_width[i], len(strip_tags(value)) + 2
                        )
                    else:
                        self._col_width[i] = max(self._col_width[i], len(value) + 2)
                except KeyError:
                    line.append(self.empty_field)
            lines.append(line)
        return lines

    def sort_data(self, sortBy, reverseSort=False):
        if self.raw_columns and sortBy in self.raw_columns:
            self.data.sort(
                key=lambda x: natural_sort_key(strip_tags(x[sortBy])),
                reverse=reverseSort,
            )
        else:
            self.data.sort(
                key=lambda x: natural_sort_key(str(x[sortBy]).strip()),
                reverse=reverseSort,
            )

    def calc_width(self):
        # Determine table widths
        for w in self._col_width:
            self._width += w
            self._width += len(self.col_sep_char)

    def build_header(self):
        headerline = ""
        for col in range(0, len(self.header)):
            if not self._ignore_col[col]:
                headerline += (
                    "<b>"
                    + escape(
                        str(" " + self.header[col] + " ").center(self._col_width[col])
                    )
                    + "</b>"
                )
            if not self._ignore_col[col + 1]:
                headerline += escape(self.col_sep_char)
        return headerline

    def build_sep(self):
        # Seperator between header and data
        header_sep_line = ""
        for i in range(0, len(self._col_width)):
            if not self._ignore_col[i]:
                for j in range(0, self._col_width[i]):
                    if self.header[i] == self.sortBy:
                        header_sep_line += self.header_sep_char_sorted
                    else:
                        header_sep_line += self.header_sep_char
                if not self._ignore_col[i + 1]:
                    header_sep_line += self.crossing_char
        return header_sep_line

    def __str__(self):
        text = ""
        if self.extra_header is not None:
            text += str(self.extra_header) + os.linesep

        # this must be before the header, because of self._ignore_col
        lines = self.build_lines()

        headerline = self.build_header()
        text += headerline + os.linesep

        header_sep_line = self.build_sep()
        text += header_sep_line + os.linesep

        for line in lines:
            line_str = ""
            for col in range(0, len(line)):
                if not self._ignore_col[col]:
                    if self.raw_columns and self.keys[col] in self.raw_columns:
                        fill_len = self._col_width[col] - len(strip_tags(line[col])) - 1
                        line_str += str(" " * fill_len + line[col] + " ")
                    else:
                        line_str += str(" " + escape(line[col]) + " ").rjust(
                            self._col_width[col]
                        )
                if not self._ignore_col[col + 1]:
                    line_str += self.col_sep_char

            text += line_str + os.linesep

        if self.extra_footer is not None:
            text += escape(str(self.extra_footer)) + os.linesep

        return text


class Position:
    """
    Class for storing postions that may be fixed in calibrated or uncalibrated space

    if self.pos_cal is set the position is fixed in calibrated space.
    if self.pos_uncal is set the position is fixed in uncalibrated space.
    """

    def __init__(self, pos, fixedInCal, cal=None):
        self.cal = cal
        self._fixedInCal = fixedInCal
        if fixedInCal:
            self._pos_cal = pos
            self._pos_uncal = None
        else:
            self._pos_cal = None
            self._pos_uncal = pos

    # pos_cal
    def _set_pos_cal(self, pos):
        if not self.fixedInCal:
            raise TypeError("Position is fixed in uncalibrated space")
        self._pos_cal = pos
        self._pos_uncal = None

    def _get_pos_cal(self):
        if self.fixedInCal:
            return self._pos_cal
        else:
            return self._Ch2E(self._pos_uncal)

    pos_cal = property(_get_pos_cal, _set_pos_cal)

    # pos_uncal
    def _set_pos_uncal(self, pos):
        if self._fixedInCal:
            raise TypeError("Position is fixed in calibrated space")
        self._pos_uncal = pos
        self._pos_cal = None

    def _get_pos_uncal(self):
        if self.fixedInCal:
            return self._E2Ch(self._pos_cal)
        else:
            return self._pos_uncal

    pos_uncal = property(_get_pos_uncal, _set_pos_uncal)

    # fixedInCal property
    def _set_fixedInCal(self, fixedInCal):
        if fixedInCal:
            self.FixInCal()
        else:
            self.FixInUncal()

    def _get_fixedInCal(self):
        return self._fixedInCal

    fixedInCal = property(_get_fixedInCal, _set_fixedInCal)

    # other functions
    def __str__(self):
        text = ""
        if self.fixedInCal:
            text += "Cal: %s" % self._pos_cal
        else:
            text += "Uncal: %s" % self._pos_uncal
        return text

    def _Ch2E(self, Ch):
        if self.cal is None:
            E = Ch
        else:
            E = self.cal.Ch2E(Ch)
        return E

    def _E2Ch(self, E):
        if self.cal is None:
            Ch = E
        else:
            Ch = self.cal.E2Ch(E)
        return Ch

    def FixInCal(self):
        """
        Fix position in calibrated space
        """
        if not self._fixedInCal:
            self._fixedInCal = True
            self.pos_cal = self._Ch2E(self._pos_uncal)

    def FixInUncal(self):
        """
        Fix position in uncalibrated space
        """
        if self._fixedInCal:
            self._fixedInCal = False
            self.pos_uncal = self._E2Ch(self._pos_cal)


class ID:
    def __init__(self, major=None, minor=None):
        if major is None:
            self.major = None
        else:
            self.major = int(major)
            if self.major < 0:
                raise ValueError("Only positive major IDs allowed")
        if minor is None:
            self.minor = None
        else:
            self.minor = int(minor)
            if self.minor < 0:
                raise ValueError("Only positive minor IDs allowed")

    def __eq__(self, other):
        if other is None:
            return False
        return (self.major, self.minor) == (other.major, other.minor)

    def __ne__(self, other):
        if other is None:
            return True
        return not ((self.major, self.minor) == (other.major, other.minor))

    def __gt__(self, other):
        try:
            if self.major == other.major:
                return self.minor > other.minor
            return self.major > other.major
        except TypeError:
            return True

    def __lt__(self, other):
        try:
            if self.major == other.major:
                return self.minor < other.minor
            return self.major < other.major
        except TypeError:
            return False

    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __hash__(self):
        # this is needed to use IDs as keys in dicts and in sets
        return hash(self.major) ^ hash(self.minor)

    def __str__(self):
        if self.major is None and self.minor is None:
            return "-"
        if self.major is None:
            return "." + str(self.minor)
        if self.minor is None:
            return str(self.major)
        return str(self.major) + "." + str(self.minor)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.__str__()}>"

    def __int__(self):
        return int(self.major)

    def __float__(self):
        if self.major is None:
            return int(self.minor) / 10.0
        if self.minor is None:
            return int(self.major)
        return int(self.major) + int(self.minor) / 10.0

    @classmethod
    def _parseSpecialID(cls, string, manager):
        if string.upper() == "NONE":
            return list()
        elif string.upper() == "NEXT":
            return [manager.nextID]
        elif string.upper() == "PREV":
            return [manager.prevID]
        elif string.upper() == "FIRST":
            return [manager.firstID]
        elif string.upper() == "LAST":
            return [manager.lastID]
        elif string.upper() == "ACTIVE":
            return [manager.activeID]
        elif string.upper() == "ALL":
            return manager.ids
        elif string.upper() == "VISIBLE":
            return list(manager.visible)
        elif string.upper() == "HIDDEN":
            return [ID for ID in manager.ids if ID not in manager.visible]
        else:
            raise ValueError

    @classmethod
    def _parseNormalID(cls, string):
        if "." in string:
            major_s, minor_s = (m for m in string.split("."))
            major = int(major_s)
            # TODO
            #            if minor_s.lower() in ("x", "y", "c"):
            #                minor = minor_s
            #            else:
            minor = int(minor_s)
        else:
            major = int(string)
            minor = None

        return ID(major, minor)

    @classmethod
    def ParseIds(cls, strings, manager, only_existent=True):
        # Normalize separators
        if not isinstance(strings, str):
            strings = ",".join(strings)
        strings = ",".join(strings.split())

        # Split string
        parts = [p for p in strings.split(",") if p]

        ids = list()
        for s in parts:
            # first deal with ranges
            if "-" in s:
                start, stop = s.split("-")
                # start
                try:
                    special = cls._parseSpecialID(start, manager)
                except ValueError:
                    start = cls._parseNormalID(start)
                except AttributeError:
                    hdtv.ui.error("Invalid key word %s" % start)
                    raise ValueError
                else:
                    if len(special) == 0:
                        start = special[0]
                    elif len(special) > 0:
                        hdtv.ui.error("Invalid ID %s" % start)
                        raise ValueError
                # stop
                try:
                    special = cls._parseSpecialID(stop, manager)
                except ValueError:
                    stop = cls._parseNormalID(stop)
                except AttributeError:
                    hdtv.ui.error("Invalid key word %s" % stop)
                    raise ValueError
                else:
                    if len(special) == 0:
                        stop = special[0]
                    elif len(special) > 0:
                        hdtv.ui.error("Invalid ID %s" % stop)
                        raise ValueError
                # fill the range
                ids.extend([i for i in manager.ids if (i >= start and i <= stop)])
            else:
                try:
                    special = cls._parseSpecialID(s, manager)
                    ids.extend(special)
                except ValueError:
                    ids.append(cls._parseNormalID(s))
                except AttributeError:
                    hdtv.ui.error("Invalid key word %s" % s)
                    raise ValueError

        # ID might be None, if e.g. activeID is None
        count = ids.count(None)
        for i in range(count):
            ids.remove(None)

        # filter non-existing ids
        valid_ids = list()
        if only_existent:
            for ID in ids:
                for mID in manager.ids:
                    if (mID.major == ID.major) and (mID.minor == ID.minor):
                        valid_ids.append(ID)
                        break  # break out of for mID in manager.ids loop

                if (
                    ID not in valid_ids
                ):  # This only works because we appended IDs above, not mIDs, because they are different instances of the ID object
                    hdtv.ui.warning("Non-existent id %s" % ID)

        else:
            valid_ids = ids

        return valid_ids


def remove_comments(string):
    """
    Removes '#' comments at the end of a line
    """
    pattern = r"(\".*?\"|\'.*?\')|(#[^\r\n]*$)"
    regex = re.compile(pattern, re.MULTILINE | re.DOTALL)

    def _replacer(match):
        if match.group(2) is not None:
            return ""
        else:
            return match.group(1)

    return regex.sub(_replacer, string)


def split_line(line):
    """
    Splits multiple commands in a single line separated by ';'
    """
    split_pattern = re.compile(r"""((?:[^;"']|"[^"]*"|'[^']*')+)""")
    return split_pattern.split(line)[1::2]


def user_save_file(filename, force=False):
    """
    Make sure filename is not in use. Offer to backup existing file
    unless force is True.
    Returns filename if successful or False if aborted.
    """
    filename = os.path.expanduser(filename)
    if not force and os.path.exists(filename):
        hdtv.ui.warning(f"This file already exists: {filename}")

        bindings = KeyBindings()

        @bindings.add("y")
        @bindings.add("Y")
        def yes(event):
            session.default_buffer.text = "y"
            event.app.exit(result=filename)

        @bindings.add("n")
        @bindings.add("N")
        def no(event):
            session.default_buffer.text = "n"
            event.app.exit(result=False)

        @bindings.add("b")
        @bindings.add("B")
        @bindings.add(Keys.Enter)
        def backup(event):
            session.default_buffer.text = "b"
            backup_file(filename)
            event.app.exit(result=filename)

        @bindings.add(Keys.Any)
        def _(event):
            pass

        session = PromptSession(
            "Replace [y/n] or backup [B] existing file? ", key_bindings=bindings
        )
        filename = session.prompt()
    return filename


def backup_file(filename, bak_ext="bak"):
    """
    Safely backup a file, using a new filename
    """
    backup_name_stem = backup_name = filename + "." + bak_ext
    for i in count():
        if not os.path.isfile(backup_name):
            break
        backup_name = backup_name_stem + "." + str(i)
    hdtv.ui.msg(f"Renamed file {filename} to {backup_name}")
    os.rename(filename, backup_name)


def open_compressed(fname, mode="rb", **kwargs):
    """
    Behaves like open(), but automatically handles compression,
    depending on the extension of the given filename.
    """
    ext = fname.split(".")[-1]
    try:
        if ext == "gz":
            import gzip

            return gzip.open(fname, mode, **kwargs)
        elif ext == "xz":
            import lzma

            return lzma.open(fname, mode, **kwargs)
        elif ext == "bz2":
            import bz2

            return bz2.open(fname, mode, **kwargs)
        return open(fname, mode, **kwargs)
    except (ImportError, AttributeError):
        raise NotImplementedError(
            "{} files are not supported. Manually use '{}' instead.".format(
                ext, "bzip2" if "bz2" else ext
            )
        )


def natural_sort_key(s, _nsre=re.compile(r"([-+]?[0-9]+\.?[0-9]*)")):
    # Split string into list of strings and floats for sorting.
    # The result of the split will always be an alternating list of strings and
    # floats (still in string form), always starting with a (possibly empty) string
    # as the first element.
    # Hence, convert all even elements to floats and all odd elements to lower case
    # strings.
    return [
        float(part) if (i % 2) else part.lower()
        for i, part in enumerate(_nsre.split(s))
    ]


class Singleton(type):
    """
    Permit no more than a single instance of a class.
    Use as a metaclass:

        class MyClass(metaclass=Singleton)
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class monkey_patch_ui:
    """Replace ROOT.HDTV.Display by a noop dummy version"""

    def __init__(self):
        import hdtv
        import hdtv.rootext.display
        import ROOT
        import hdtv.dummy

        self._orig = ROOT.HDTV.Display
        ROOT.HDTV.Display = hdtv.dummy

    def __enter__(self):
        pass

    def __exit__(self, *args, **kws):
        import ROOT

        ROOT.HDTV.Display = self._orig


@contextlib.contextmanager
def temp_seed(seed: Optional[int] = None):
    """
    Temporarily set numpy seed. Restore old state of random
    number generator afterwards.

    Args:
        seed: Temporary seed for numpy
    """
    state = np.random.get_state() if seed else None
    np.random.seed(seed)
    try:
        yield
    finally:
        if state:
            np.random.set_state(state)


def SplitCmdline(s):
    """
    Split a string, handling escaped whitespace.
    Essentially our own version of shlex.split, but with only double
    quotes accepted as quotes.

    Returns:
        List of command fragments
        Suffix removed from last fragment (quotes)
    """
    try:
        try:
            lex = shlex.shlex(s, posix=True)
            lex.quotes = r'"'
            lex.commenters = "#"
            lex.whitespace_split = True
            return list(lex), ""
        except ValueError:
            lex = shlex.shlex(s + '"', posix=True)
            lex.quotes = r'"'
            lex.commenters = "#"
            lex.whitespace_split = True
            return list(lex), '"'
    except ValueError:
        return []


def SplitCmdlines(s):
    """
    Split line into multiple commands separated by ';'.
    """
    cmds = re.findall(r'(?:[^;"]|"(?:|[^"])*(?:"|$))+', s)
    last_suffix = ""
    segs = []
    for cmd in cmds:
        seg, last_suffix = SplitCmdline(cmd)
        segs.append(seg)
    return segs, last_suffix


class LockViewport:
    def __init__(self, viewport):
        """
        Context manager to lock viewport
        """
        self.viewport = viewport

    def __enter__(self):
        if self.viewport:
            self.viewport.LockUpdate()

    def __exit__(self, *a):
        if self.viewport:
            self.viewport.UnlockUpdate()



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import ROOT

import hdtv.peakmodels
import hdtv.backgroundmodels
from hdtv.util import Pairs


class Fitter:
    """
    PeakModel independent part of the Interface to the C++ Fitter
    The parts that depend on a special peak model can be found in peak.py.
    """

    def __init__(self, peakModel, backgroundModel):
        self.SetPeakModel(peakModel)
        self.SetBackgroundModel(backgroundModel)
        self.peakFitter = None
        self.bgFitter = None

    @property
    def params(self):
        params = ["background"]
        for param, valid_status in self.peakModel.fValidParStatus.items():
            if not "calculated" in valid_status:
                params.append(param)
        for param in self.peakModel.fValidOptStatus.keys():
            params.append(param)
        return params

    def __getattr__(self, name):
        # Look in peakModel for unknown attributes
        return getattr(self.peakModel, name)

    def FitBackground(self, spec, backgrounds=Pairs()):
        """
        Create Background Fitter object and do the background fit
        """
        # create fitter
        self.bgFitter = self.backgroundModel.GetFitter(
            integrate=self.peakModel.GetOption("integrate"),
            likelihood=self.peakModel.GetOption("likelihood"),
            nparams=self.backgroundModel.fParStatus["nparams"],
            nbg=len(backgrounds),
        )
        if self.bgFitter is None:
            msg = "Background model %s needs at least %i background regions to execute a fit. Found %i.".format(
                self.backgroundModel.name,
                self.backgroundModel.requiredBgRegions,
                len(backgrounds),
            )
            raise ValueError(msg)
        else:
            for bg in backgrounds:
                self.bgFitter.AddRegion(bg[0], bg[1])
            # do the background fit
            self.bgFitter.Fit(spec.hist.hist)

    def RestoreBackground(self, backgrounds=Pairs(), params=list(), chisquare=0.0):
        """
        Create Background Fitter object and
        restore the background polynom from coeffs
        """
        self.bgFitter = self.backgroundModel.GetFitter(
            integrate=self.peakModel.GetOption("integrate"),
            likelihood=self.peakModel.GetOption("likelihood"),
            nparams=len(params),
            nbg=len(backgrounds),
        )
        # restore the fitter
        valueArray = ROOT.TArrayD(len(params))
        errorArray = ROOT.TArrayD(len(params))
        for i, param in enumerate(params):
            valueArray[i] = param.nominal_value
            errorArray[i] = param.std_dev
        self.bgFitter.Restore(valueArray, errorArray, chisquare)

    def FitPeaks(self, spec, region=Pairs(), peaklist=list()):
        """
        Create the Peak Fitter object and do the peak fit
        """
        # create the fitter
        self.peakFitter = self.peakModel.GetFitter(region, peaklist, spec.cal)
        # Do the peak fit
        if self.bgFitter:
            # external background
            self.peakFitter.Fit(spec.hist.hist, self.bgFitter)
        else:
            # internal background
            self.peakFitter.Fit(
                spec.hist.hist, self.backgroundModel.fParStatus["nparams"]
            )

    def RestorePeaks(
        self, cal=None, region=Pairs(), peaks=list(), chisquare=0.0, coeffs=list()
    ):
        """
        Create the Peak Fitter object and
        restore all peaks
        """
        # create the fitter
        peaklist = [p.pos.nominal_value for p in peaks]
        self.peakFitter = self.peakModel.GetFitter(region, peaklist, cal)
        # restore first the fitter and afterwards the peaks
        if self.bgFitter:
            # external background
            self.peakFitter.Restore(self.bgFitter, chisquare)
        else:
            # internal background
            values = ROOT.TArrayD(self.backgroundModel.fParStatus["nparams"])
            errors = ROOT.TArrayD(self.backgroundModel.fParStatus["nparams"])
            for i in range(0, self.backgroundModel.fParStatus["nparams"]):
                values[i] = coeffs[i].nominal_value
                errors[i] = coeffs[i].std_dev
            self.peakFitter.Restore(values, errors, chisquare)
        if not len(peaks) == self.peakFitter.GetNumPeaks():
            raise RuntimeError("Number of peaks does not match")
        for i, peak in enumerate(peaks):
            self.peakModel.RestoreParams(peak, self.peakFitter.GetPeak(i))

    def SetPeakModel(self, model):
        """
        Sets the peak model to be used for fitting.
        Model can be either a string, in which case it is used as a key into
        the gPeakModels dictionary, or a PeakModel object.
        """
        if isinstance(model, str):
            model = hdtv.peakmodels.PeakModels[model]
        self.peakModel = model()
        self.peakFitter = None

    def SetBackgroundModel(self, model):
        """
        Sets the background model to be used for fitting.
        Model can be either a string, in which case it is used as a key into
        the gPeakModels dictionary, or a BackgroundModel object.
        """
        if isinstance(model, str):
            model = hdtv.backgroundmodels.BackgroundModels[model]
        self.backgroundModel = model()
        self.backgroundFitter = None

    def SetParameter(self, parname, status):
        """
        Sets the parameter status for fitting.
        """
        if parname == "background":
            self.backgroundModel.SetParameter("nparams", status)
        else:
            # all other parnames are treated in the peakmodel
            self.peakModel.SetParameter(parname, status)

    def __copy__(self):
        """
        Create a copy of this fitter
        This also copies the status of the corresponding peakModel,
        and hence the status of the fit parameters and options.
        """
        new = Fitter(self.peakModel.name, self.backgroundModel.name)
        new.peakModel.fParStatus = self.peakModel.fParStatus.copy()
        new.peakModel.fOptStatus = self.peakModel.fOptStatus.copy()
        new.backgroundModel.fParStatus = self.backgroundModel.fParStatus.copy()
        return new

    def __eq__(self, other):
        return (
            isinstance(other, Fitter)
            and self.peakModel.name == other.peakModel.name
            and self.backgroundModel.name == other.backgroundModel.name
            and self.peakModel.fParStatus == other.peakModel.fParStatus
            and self.peakModel.fOptStatus == other.peakModel.fOptStatus
            and self.backgroundModel.fParStatus == other.backgroundModel.fParStatus
        )



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2019  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA


import colorsys
import math
import ROOT
from hdtv.hsluv import hsluv_to_rgb, rgb_to_hsluv


# some default colors
zoom = ROOT.kWhite
region = ROOT.kBlue - 4
peak = ROOT.kViolet - 4
bg = ROOT.kGreen - 4
cut = ROOT.kYellow - 4

active_luminance = 76
nonactive_luminance = 58
golden_angle = 180 * (3 - math.sqrt(5))


def ColorForID(ID, active=False):
    try:
        ID = ID.split(".")
        ID = int(ID[0])
    except BaseException:
        pass
    hue = (ID * golden_angle) % 360
    saturation = 100
    luminance = active_luminance if active else nonactive_luminance
    r, g, b = hsluv_to_rgb([hue, saturation, luminance])
    return ROOT.TColor.GetColor(r, g, b)


default = ColorForID(0)


def Highlight(color, active=True):
    """
    Manipulates the color in HSV room to achieve highlighting
     - The hue value runs from 0 to 360.
     - The saturation is the degree of strength or purity and is from 0 to 1.
       Purity is how much white is added to the color, so S=1 makes the purest
       color (no white).
     - Brightness value also ranges from 0 to 1, where 0 is black.
    """
    if color is None:
        color = default
    # you can not highlight white and black
    if color == 10 or color == 0:
        return color
    color = ROOT.gROOT.GetColor(color)
    if not color:
        # FIXME
        raise RuntimeError
    color_r = ROOT.TColor.GetRed(color)
    color_g = ROOT.TColor.GetGreen(color)
    color_b = ROOT.TColor.GetBlue(color)
    hue, saturation, luminance = rgb_to_hsluv([color_r, color_g, color_b])
    saturation = 100
    luminance = active_luminance if active else nonactive_luminance
    r, g, b = hsluv_to_rgb([hue, saturation, luminance])
    return ROOT.TColor.GetColor(r, g, b)


def GetRGB(color):
    color = ROOT.gROOT.GetColor(color)
    if not color:
        # FIXME
        raise RuntimeError
    r = ROOT.TColor.GetRed(color)
    g = ROOT.TColor.GetGreen(color)
    b = ROOT.TColor.GetBlue(color)
    return (r, g, b)



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import array
import string
from math import sqrt
from uncertainties import ufloat

from hdtv.util import TxtFile, Pairs
import hdtv.ui
from ROOT import TF1, TF2, TGraphErrors, TVirtualFitter


class _Efficiency:
    def __init__(self, num_pars=0, pars=None, norm=True):
        pars = pars or []

        self._numPars = num_pars
        self.parameter = pars
        self.fCov = [
            [None for j in range(self._numPars)] for i in range(self._numPars)
        ]  # Simple matrix replacement
        self._dEff_dP = list()
        self.TGraph = TGraphErrors()

        # Normalization factors
        self._doNorm = norm
        self.norm = 1.0
        self.TF1.FixParameter(0, self.norm)  # Normalization
        self.TF1.SetRange(0, 10000)  # Default range for efficiency function

        self._fitInput = Pairs(lambda x: ufloat(x, 0))

        #         if self.parameter: # Parameters were given
        #             map(lambda i: self.TF1.SetParameter(i + 1, self.parameter[i]), range(1, len(pars))) # Set initial parameters
        #         else:
        #             self.parameter = [None for i in range(1, self._numPars + 1)]
        #
        self.TF1.SetParName(0, "N")  # Normalization

        for i in range(0, num_pars):
            self._dEff_dP.append(None)
            if num_pars <= len(string.ascii_lowercase):
                self.TF1.SetParName(i + 1, string.ascii_lowercase[i])

    def _getParameter(self):
        """
        Get parameter of efficiency function
        """
        pars = list()
        for i in range(self._numPars):
            pars.append(self.TF1.GetParameter(i))

        return pars

    def _setParameter(self, pars):
        """
        Set parameter for efficiency function
        """
        for i in range(self._numPars):
            try:
                self.TF1.SetParameter(i, pars[i])
            except IndexError:
                self.TF1.SetParameter(i, 0)

    parameter = property(_getParameter, _setParameter)

    def __call__(self, E):
        value = self.value(E)
        try:
            error = self.error(E)
        except TypeError:
            error = None
        return ufloat(value, error)

    def _set_fitInput(self, fitPairs):

        self._fitInput = fitPairs

        for i in range(len(self._fitInput)):
            p = self._fitInput[i]
            try:
                e_nominal_value = p[0].nominal_value
                e_std_dev = p[0].std_dev
            except AttributeError:
                e_nominal_value = float(p[0])
                e_std_dev = 0.0

            try:
                eff_nominal_value = p[1].nominal_value
                eff_std_dev = p[1].std_dev
            except AttributeError:
                eff_nominal_value = float(p[1])
                eff_std_dev = 0.0

            self.TGraph.SetPoint(i, e_nominal_value, eff_nominal_value)
            self.TGraph.SetPointError(i, e_std_dev, eff_std_dev)

    def _get_fitInput(self):
        return self._fitInput

    fitInput = property(_get_fitInput, _set_fitInput)

    def fit(self, fitPairs=None, quiet=True):
        """
        Fit efficiency curve to values given by 'fitPairs' which should be a list
        of energy<->efficiency pairs. (See hdtv.util.Pairs())

        'energies' and 'efficiencies' may be a list of ufloats
        """
        # TODO: Unify this with the energy calibration fitter
        if fitPairs is not None:
            # self.fitInput = fitPairs
            self._fitInput = fitPairs

        E = array.array("d")
        delta_E = array.array("d")
        eff = array.array("d")
        delta_eff = array.array("d")
        EN = array.array("d")
        effN = array.array("d")

        #        map(energies.append(self.fitInput[0]), self.fitInput)
        #        map(efficiencies.append(self.fitInput[1]), self.fitInput)
        hasXerrors = False
        # Convert energies to array needed by ROOT
        try:
            list(map(lambda x: E.append(x[0].nominal_value), self._fitInput))
            list(map(lambda x: delta_E.append(x[0].std_dev), self._fitInput))
            list(map(lambda x: EN.append(0.0), self._fitInput))
            hasXerrors = True
        except AttributeError:  # energies does not seem to be ufloat list
            list(map(lambda x: E.append(x[0]), self._fitInput))
            list(map(lambda x: delta_E.append(0.0), self._fitInput))

        # Convert efficiencies to array needed by ROOT
        try:
            list(map(lambda x: eff.append(x[1].nominal_value), self._fitInput))
            list(map(lambda x: delta_eff.append(x[1].std_dev), self._fitInput))
            list(map(lambda x: effN.append(0.0), self._fitInput))
        except AttributeError:  # energies does not seem to be ufloat list
            list(map(lambda x: eff.append(x[1]), self._fitInput))
            list(map(lambda x: delta_eff.append(0.0), self._fitInput))

        # if fit has errors we first fit without errors to get good initial values
        # if hasXerrors == True:
        # print "Fit parameter without errors included:"
        # self.TGraphWithoutErrors = TGraphErrors(len(E), E, eff, EN, effN)
        # fitWithoutErrors = self.TGraphWithoutErrors.Fit(self.id, "SF")

        hdtv.ui.msg("Fit parameter with errors included:")

        # Preliminary normalization
        #        if self._doNorm:
        #            self.norm = 1 / max(efficiencies)
        #            for i in range(len(eff)):
        #                eff[i] *= self.norm
        #                delta_eff[i] *= self.norm

        # self.TF1.SetRange(0, max(E) * 1.1)
        # self.TF1.SetParameter(0, 1) # Unset normalization for fitting

        self.TGraph = TGraphErrors(len(E), E, eff, delta_E, delta_eff)

        # Do the fit
        fitopts = "0"  # Do not plot

        if hasXerrors:
            # We must use the iterative fitter (minuit) to take x errors
            # into account.
            fitopts += "F"
            hdtv.ui.info(
                "switching to non-linear fitter (minuit) for x error weighting"
            )
        if quiet:
            fitopts += "Q"

        fitopts += (
            "S"  # Additional fitinfo returned needed for ROOT5.26 workaround below
        )
        fitreturn = self.TGraph.Fit(self.id, fitopts)

        try:
            # Workaround for checking the fitstatus in ROOT 5.26 (TFitResultPtr
            # does not cast properly to int)
            fitstatus = fitreturn.Get().Status()
        except AttributeError:  # This is for ROOT <= 5.24, where fit returns an int
            fitstatus = int(fitreturn)

        if fitstatus != 0:
            # raise RuntimeError, "Fit failed"
            hdtv.ui.msg("Fit failed")

        #         # Final normalization
        #         if self._doNorm:
        #             self.normalize()

        # Get parameter
        for i in range(self._numPars):
            self.parameter[i] = self.TF1.GetParameter(i)

        # Get covariance matrix
        tvf = TVirtualFitter.GetFitter()
        ##        cov = tvf.GetCovarianceMatrix()
        for i in range(0, self._numPars):
            for j in range(0, self._numPars):
                self.fCov[i][j] = tvf.GetCovarianceMatrixElement(i, j)
        ##                 self.fCov[i][j] = cov[i * self._numPars + j]

        return self.parameter

    def normalize(self):
        # Normalize the efficiency funtion
        try:
            self.norm = 1.0 / self.TF1.GetMaximum(0.0, 0.0)
        except ZeroDivisionError:
            self.norm = 1.0

        self.TF1.SetParameter(0, self.norm)
        normfunc = TF2("norm_" + hex(id(self)), "[0]*y")
        normfunc.SetParameter(0, self.norm)
        self.TGraph.Apply(normfunc)

    def value(self, E):
        try:
            value = E.nominal_value
        except AttributeError:
            value = E

        return self.TF1.Eval(value)

    def error(self, E):
        """
        Calculate error using the covariance matrix via:

          delta_Eff = sqrt((dEff_dP[0], dEff_dP[1], ... dEff_dP[num_pars]) x cov x (dEff_dP[0], dEff_dP[1], ... dEff_dP[num_pars]))

        """
        try:
            value = E.nominal_value
        except AttributeError:
            value = E

        if not self.fCov or (len(self.fCov) != self._numPars):
            raise ValueError("Incorrect size of covariance matrix")

        res = 0.0

        # Do matrix multiplication
        for i in range(0, self._numPars):
            tmp = 0.0
            for j in range(0, self._numPars):
                tmp += self._dEff_dP[j](value, self.parameter) * self.fCov[i][j]

            res += self._dEff_dP[i](value, self.parameter) * tmp

        return sqrt(res)

    def loadPar(self, parfile):
        """
        Read parameter from file
        """
        vals = []

        file = TxtFile(parfile)
        file.read()

        for line in file.lines:
            vals.append(float(line))

        if len(vals) != self._numPars:
            raise RuntimeError("Incorrect number of parameters found in file")

        self.parameter = vals
        if self._doNorm:
            self.normalize()

    def loadCov(self, covfile):
        """
        Load covariance matrix from file
        """

        vals = []

        file = TxtFile(covfile)
        file.read()

        for line in file.lines:
            val_row = [float(s) for s in line.split()]
            if len(val_row) != self._numPars:
                raise RuntimeError("Incorrect format of parameter error file")
            vals.append(val_row)

        if len(vals) != self._numPars:
            raise RuntimeError("Incorrect format of parameter error file")

        self.fCov = vals

    def load(self, parfile, covfile=None):
        """
        Read parameter and covariance matrix from file
        """
        self.loadPar(parfile)

        if covfile:
            self.loadCov(covfile)

    def savePar(self, parfile):
        """
        Save parameter to file
        """
        file = TxtFile(parfile, "w")

        for p in self.parameter:
            file.lines.append(str(p))

        file.write()

    def saveCov(self, covfile):
        """
        Save covariance matrix to file
        """
        file = TxtFile(covfile, "w")

        for i in range(0, self._numPars):
            line = ""
            for j in range(0, self._numPars):
                line += str(self.fCov[i][j]) + " "
            file.lines.append(line.strip())

        file.write()

    def save(self, parfile, covfile=None):
        """
        Save parameter and covariance matrix to files
        """
        # Write paramter
        self.savePar(parfile)

        # Write covariance matrix
        if covfile is not None:
            self.saveCov(covfile)



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import ROOT
import os
import re
import glob
import copy

import hdtv.cmdline
import hdtv.color
import hdtv.cal
import hdtv.options
import hdtv.util
import hdtv.ui

from hdtv.spectrum import Spectrum
from hdtv.histogram import FileHistogram
from hdtv.specreader import SpecReaderError
from hdtv.util import LockViewport


class SpecInterface:
    """
    User interface to work with 1-d spectra
    """

    def __init__(self, spectra):
        hdtv.ui.debug("Loaded user interface for working with 1-d spectra")

        self.spectra = spectra
        self.window = spectra.window
        self.caldict = spectra.caldict

        # tv commands
        self.tv = TvSpecInterface(self)

        # good to have as well...
        if self.window:
            self.window.AddHotkey(ROOT.kKey_PageUp, self.spectra.ShowPrev)
            self.window.AddHotkey(ROOT.kKey_PageDown, self.spectra.ShowNext)

            # register common tv hotkeys
            self.window.AddHotkey([ROOT.kKey_N, ROOT.kKey_p], self.spectra.ShowPrev)
            self.window.AddHotkey([ROOT.kKey_N, ROOT.kKey_n], self.spectra.ShowNext)
            self.window.AddHotkey(ROOT.kKey_Equal, self.spectra.RefreshAll)
            self.window.AddHotkey(ROOT.kKey_t, self.spectra.RefreshVisible)
            self.window.AddHotkey(
                ROOT.kKey_n,
                lambda: self.window.EnterEditMode(
                    prompt="Show spectrum: ", handler=self._HotkeyShow
                ),
            )
            self.window.AddHotkey(
                ROOT.kKey_a,
                lambda: self.window.EnterEditMode(
                    prompt="Activate spectrum: ", handler=self._HotkeyActivate
                ),
            )

        # Register configuration variables
        self.LoadSpectraNaturalSort = hdtv.options.Option(
            default=True,
            parse=hdtv.options.parse_bool,
        )
        hdtv.options.RegisterOption(
            "spec.get.sort_natural", self.LoadSpectraNaturalSort
        )

    def _HotkeyShow(self, arg):
        """
        ShowObjects wrapper for use with Hotkey
        """
        try:
            ids = hdtv.util.ID.ParseIds(arg, self.spectra)
            if len(ids) == 0:
                self.spectra.HideAll()
            else:
                self.spectra.ShowObjects(ids)
        except ValueError:
            if self.window:
                self.window.viewport.SetStatusText(
                    "Invalid spectrum identifier: %s" % arg
                )

    def _HotkeyActivate(self, arg):
        """
        ActivateObject wrapper for use with Hotkey
        """
        try:
            ids = hdtv.util.ID.ParseIds(arg, self.spectra)
            if len(ids) > 1:
                if self.window:
                    self.window.viewport.SetStatusText(
                        "Cannot activate more than one spectrum"
                    )
            elif len(ids) == 0:  # Deactivate
                oldactive = self.spectra.activeID
                self.spectra.ActivateObject(None)
                if self.window:
                    self.window.viewport.SetStatusText(
                        "Deactivated spectrum %s" % oldactive
                    )
            else:
                ID = ids[0]
                self.spectra.ActivateObject(ID)
                if self.window:
                    self.window.viewport.SetStatusText(
                        "Activated spectrum %s" % self.spectra.activeID
                    )
        except ValueError:
            if self.window:
                self.window.viewport.SetStatusText("Invalid id: %s" % arg)

    def LoadSpectra(self, patterns, ID=None):
        """
        Load spectra from files matching patterns.

        If ID is specified, the spectrum is stored with id ID, possibly
        replacing a spectrum that was there before.

        Returns:
            A list of the loaded spectra
        """
        with LockViewport(self.window.viewport if self.window else None):
            # only one filename is given
            if isinstance(patterns, str):
                patterns = [patterns]

            if ID is not None and len(patterns) > 1:
                raise hdtv.cmdline.HDTVCommandError(
                    "If you specify an ID, you can only give one pattern"
                )

            loaded = []
            for p in patterns:
                # put fmt if available
                p = p.rsplit("'", 1)
                if len(p) == 1 or not p[1]:
                    (fpat, fmt) = (p[0], None)
                else:
                    (fpat, fmt) = p

                files = glob.glob(os.path.expanduser(fpat))

                if len(files) == 0:
                    hdtv.ui.warning("%s: no such file" % fpat)
                elif ID is not None and len(files) > 1:
                    raise hdtv.cmdline.HDTVCommandAbort(
                        "pattern %s is ambiguous and you specified an ID" % fpat
                    )
                    break

                if self.LoadSpectraNaturalSort.Get():
                    files.sort(key=hdtv.util.natural_sort_key)
                else:
                    files.sort()

                for fname in files:
                    try:
                        # Create spectrum object
                        spec = Spectrum(FileHistogram(fname, fmt))
                    except (OSError, SpecReaderError):
                        hdtv.ui.warning(f"Could not load {fname}'{fmt}")
                    else:
                        sid = self.spectra.Insert(spec, ID)
                        spec.color = hdtv.color.ColorForID(sid.major)
                        try:
                            spec.cal = self.spectra.caldict[spec.name]
                        except KeyError:
                            pass
                        loaded.append(spec)
                        if fmt is None:
                            hdtv.ui.msg(f"Loaded {fname} into {sid}")
                        else:
                            hdtv.ui.msg(f"Loaded {fname}'{fmt} into {sid}")

            if loaded:
                # activate last loaded spectrum
                self.spectra.ActivateObject(loaded[-1].ID)
            # Expand window if it is the only spectrum
            if len(self.spectra) == 1:
                if self.window:
                    self.window.Expand()
            return loaded

    def ListSpectra(self, visible=False):
        """
        Create a list of all spectra (for printing)
        """
        spectra = list()
        params = ["ID", "stat", "name", "fits"]

        for (ID, obj) in self.spectra.dict.items():
            if visible and (ID not in self.spectra.visible):
                continue

            thisspec = dict()

            status = ""
            if ID == self.spectra.activeID:
                status += "A"
            if ID in self.spectra.visible:
                status += "V"

            if ID == self.spectra.activeID:
                color = self.spectra.dict[ID]._activeColor
            else:
                color = self.spectra.dict[ID]._passiveColor
            r, g, b = hdtv.color.GetRGB(color)
            r, g, b = int(255 * r), int(255 * g), int(255 * b)
            thisspec["ID"] = f'<style fg="#{r:02x}{g:02x}{b:02x}">{ID}</style>'
            thisspec["stat"] = status
            thisspec["name"] = self.spectra.dict[ID].name
            thisspec["fits"] = len(self.spectra.dict[ID].dict)
            spectra.append(thisspec)

        table = hdtv.util.Table(spectra, params, sortBy="ID", raw_columns=["ID"])
        hdtv.ui.msg(html=str(table), end="")

    def CopySpectrum(self, ID, copyTo=None):
        """
        Copy spectrum

        Return ID of new spectrum
        """
        if copyTo is None:
            copyTo = self.spectra.GetFreeID()
        # get copy of underlying histogramm
        hist = copy.copy(self.spectra.dict[ID].hist)
        # create new spectrum object
        spec = Spectrum(hist)
        spec.color = hdtv.color.ColorForID(copyTo.major)
        spec.typeStr = "spectrum, copy"
        spec.name = "%s (copy)" % spec.name
        if hist.cal:
            cal = copy.copy(hist.cal)
            self.caldict[spec.name] = cal
            spec.cal = cal
        sid = self.spectra.Insert(spec, copyTo)
        hdtv.ui.msg("Copied spectrum " + str(ID) + " to " + str(sid))


class TvSpecInterface:
    """
    TV style commands for the spectrum interface.
    """

    def __init__(self, specInterface):
        self.opt = dict()
        self.specIf = specInterface
        self.spectra = self.specIf.spectra

        # spectrum commands
        prog = "spectrum get"
        description = "Load a spectrum from a file (using libmfile)"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default=None,
            help="id for loaded spectrum",
        )
        parser.add_argument("pattern", nargs="+")
        hdtv.cmdline.AddCommand(
            prog, self.SpectrumGet, level=0, fileargs=True, parser=parser
        )
        # the spectrum get command is registered with level=0,
        # this allows "spectrum get" to be abbreviated as "spectrum", register
        # all other commands starting with spectrum with default or higher
        # priority

        prog = "spectrum list"
        description = "List all spectra"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-v",
            "--visible",
            action="store_true",
            default=False,
            help="list only visible (and active) spectra",
        )
        hdtv.cmdline.AddCommand(prog, self.SpectrumList, parser=parser)

        prog = "spectrum delete"
        description = "Delete one or multiple spectra"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("specid", nargs="*", help="id of spectrum to delete")
        hdtv.cmdline.AddCommand(prog, self.SpectrumDelete, parser=parser)

        prog = "spectrum activate"
        description = (
            "Activate a single spetrum. If hidden, the spectrum is set to visible."
        )
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("specid", nargs="*", help="id of spectrum to activate")
        hdtv.cmdline.AddCommand(prog, self.SpectrumActivate, parser=parser)

        prog = "spectrum show"
        description = "Set one or multiple spectra to visible"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "specid", nargs="*", help="id (or all, shown) of spectrum to update"
        )
        hdtv.cmdline.AddCommand(prog, self.SpectrumShow, parser=parser)

        prog = "spectrum hide"
        description = "Set one or multiple spectra to hidden"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "specid", nargs="*", help="id (or all, shown) of spectrum to update"
        )
        hdtv.cmdline.AddCommand(prog, self.SpectrumHide, level=2, parser=parser)

        prog = "spectrum info"
        description = "Display information associated with one or multiple spectra"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("specid", nargs="*", help="id of spectrum to update")
        hdtv.cmdline.AddCommand(prog, self.SpectrumInfo, parser=parser)

        prog = "spectrum update"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog)
        parser.add_argument(
            "specid", nargs="*", help="id (or all, shown) of spectrum to update"
        )
        hdtv.cmdline.AddCommand(prog, self.SpectrumUpdate, parser=parser)

        prog = "spectrum write"
        description = "Write a single spectrum to the filesystem (using libmfile)"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("filename", help="filename of output file")
        parser.add_argument(
            "format", help="format of spectrum file (e.g. txt for simple text file)"
        )
        parser.add_argument(
            "specid", nargs="?", default=None, help="id of spectrum to write"
        )
        parser.add_argument(
            "-F",
            "--force",
            action="store_true",
            default=False,
            help="overwrite existing files without asking",
        )
        hdtv.cmdline.AddCommand(prog, self.SpectrumWrite, parser=parser)

        prog = "spectrum normalize"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog)
        parser.add_argument(
            "specid", nargs="*", default=None, help="id of spectrum to normalize"
        )
        parser.add_argument("norm", type=float, help="norm of spectrum")
        hdtv.cmdline.AddCommand(prog, self.SpectrumNormalization, parser=parser)

        prog = "spectrum rebin"
        description = "Rebin the spectrum by grouping neighboring bins. The calibration of the spectrum is updated accordingly."
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "specid", nargs="*", default=None, help="id of spectrum to rebin"
        )
        parser.add_argument("ngroup", type=int, help="group n bins for rebinning")
        parser.add_argument(
            "-c",
            "--calibrate",
            action="store_true",
            help="Calibrate the spectrum after rebinning. If this switch is not used, peak positions will be shifted from pos to 1/ngroup*pos.",
        )
        hdtv.cmdline.AddCommand(
            prog, self.SpectrumRebin, level=2, fileargs=False, parser=parser
        )

        self.opt["binning"] = hdtv.options.Option(
            default="tv", parse=hdtv.options.parse_choices(["root", "tv"])
        )
        hdtv.options.RegisterOption("spec.binning", self.opt["binning"])

        prog = "spectrum calbin"
        description = "Align binning to energy calibration of spectrum."
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "specid", nargs="*", default=None, help="id of spectrum to calbin"
        )
        parser.add_argument(
            "--spline-order",
            "-k",
            type=int,
            default=3,
            metavar="[1-5]",
            help="Order of the spline interpolation (default: %(default)s)",
        )
        parser.add_argument(
            "-d", "--deterministic", action="store_true", help="Rebin deterministically"
        )
        parser.add_argument("--seed", "-s", type=int, help="Set random number seed")
        parser.add_argument(
            "--binsize",
            "-b",
            type=float,
            default=1.0,
            help="Size of each bin after rebinning (default: %(default)s)",
        )
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--root-binning",
            "-r",
            action="store_true",
            help="Lower edge of first bin is aligned to 0.",
        )
        group.add_argument(
            "--tv-binning",
            "-t",
            action="store_true",
            help="Center of first bin is aligned to 0.",
        )
        hdtv.cmdline.AddCommand(
            prog, self.SpectrumCalbin, level=2, fileargs=False, parser=parser
        )

        prog = "spectrum resample"
        description = "Vary the content of each bin within its uncertainty, creating a new random sample of the underlying probability distribution."
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "specid", nargs="*", default=None, help="id of spectrum to resample"
        )
        parser.add_argument(
            "--distribution",
            "-d",
            type=str,
            default="poisson",
            help="Specify a random distribution function (default: %(default)s)",
        )
        parser.add_argument("--seed", "-s", type=int, help="Set random number seed")
        hdtv.cmdline.AddCommand(
            prog,
            self.SpectrumResample,
            level=2,
            fileargs=False,
            completer=self.RandomModelCompleter,
            parser=parser,
        )

        prog = "spectrum add"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog)
        parser.add_argument(
            "-n",
            "--normalize",
            action="store_true",
            help="normalize <target-id> by dividing through number of added spectra afterwards",
        )
        parser.add_argument(
            "targetid",
            metavar="target-id",
            help="where to place the resulting spectrum",
        )
        parser.add_argument("specid", nargs="*", help="ids of spectra to add up")
        hdtv.cmdline.AddCommand(
            prog, self.SpectrumAdd, level=2, fileargs=False, parser=parser
        )

        prog = "spectrum subtract"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog)
        parser.add_argument(
            "targetid",
            metavar="target-id",
            help="where to place the resulting spectrum",
        )
        parser.add_argument("specid", nargs="*", help="ids of spectra to subtract")
        hdtv.cmdline.AddCommand(
            prog, self.SpectrumSub, level=2, fileargs=False, parser=parser
        )

        prog = "spectrum multiply"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog)
        parser.add_argument(
            "specid", nargs="*", default=None, help="id of spectrum to multiply"
        )
        parser.add_argument("factor", type=float, help="multiplication factor")
        hdtv.cmdline.AddCommand(
            prog, self.SpectrumMultiply, level=2, fileargs=False, parser=parser
        )

        prog = "spectrum copy"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog)
        parser.add_argument(
            "specid", nargs="*", default=None, help="id of spectrum to copy"
        )
        parser.add_argument(
            "-s", "--spectrum", action="store", default=None, help="Target spectrum id"
        )
        hdtv.cmdline.AddCommand(
            prog, self.SpectrumCopy, level=2, fileargs=False, parser=parser
        )

        prog = "spectrum name"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog)
        parser.add_argument(
            "specid", nargs="?", default=None, help="id of spectrum to name"
        )
        parser.add_argument("name", help="name of spectrum")
        hdtv.cmdline.AddCommand(
            prog, self.SpectrumName, level=2, fileargs=False, parser=parser
        )

    def SpectrumList(self, args):
        """
        Print a list of spectra
        """
        self.specIf.ListSpectra(visible=args.visible)

    def SpectrumGet(self, args):
        """
        Load Spectra from files
        """
        if args.spectrum is not None:
            try:
                ids = hdtv.util.ID.ParseIds(
                    args.spectrum, self.spectra, only_existent=False
                )
                if len(ids) > 1:
                    raise hdtv.cmdline.HDTVCommandError("More than one ID given")
                ID = ids[0]
            except ValueError as msg:
                raise hdtv.cmdline.HDTVCommandError("Invalid ID: %s" % msg)
        else:
            ID = None
        self.specIf.LoadSpectra(patterns=args.pattern, ID=ID)

    def SpectrumDelete(self, args):
        """
        Deletes spectra
        """
        ids = hdtv.util.ID.ParseIds(args.specid, self.spectra)

        if len(ids) == 0:
            hdtv.ui.warning("Nothing to do")
            return
        for ID in ids:
            self.spectra.Pop(ID)

    def SpectrumActivate(self, args):
        """
        Activate one spectrum
        """
        ids = hdtv.util.ID.ParseIds(args.specid, self.spectra)

        if len(ids) > 1:
            raise hdtv.cmdline.HDTVCommandError("Can only activate one spectrum")
        elif len(ids) == 0:
            self.spectra.ActivateObject(None)
        else:
            self.spectra.ActivateObject(min(ids))

    def SpectrumCopy(self, args):
        """
        Copy spectra
        """
        ids = hdtv.util.ID.ParseIds(args.specid, self.spectra)

        if len(ids) == 0:
            hdtv.ui.warning("Nothing to do")
            return
        targetids = list()
        if args.spectrum is not None:
            targetids = hdtv.util.ID.ParseIds(
                args.spectrum, self.spectra, only_existent=False
            )
            targetids.sort()
        if len(targetids) == 0:
            targetids = [None for i in range(0, len(ids))]
        elif len(targetids) == 1:  # Only start ID is given
            startID = targetids[0]
            targetids = [
                hdtv.util.ID(i) for i in range(startID.major, startID.major + len(ids))
            ]
            targetids.sort()
        elif len(targetids) != len(ids):
            raise hdtv.cmdline.HDTVCommandError(
                "Number of target ids does not match number of ids to copy"
            )
        for i in range(0, len(ids)):
            try:
                self.specIf.CopySpectrum(ids[i], copyTo=targetids[i])
            except KeyError:
                raise hdtv.cmdline.HDTVCommandError("No such spectrum: " + str(ids[i]))

    def SpectrumAdd(self, args):
        """
        Add spectra (spec1 + spec2, ...)
        """
        # FIXME: Properly separate targetid, specid
        ids = hdtv.util.ID.ParseIds(
            [args.targetid] + args.specid, self.spectra, only_existent=False
        )

        if len(ids) == 0:
            hdtv.ui.warning("Nothing to do")
            return

        addTo = ids[0]
        # if addTo is a new spectrum
        if addTo not in list(self.spectra.dict.keys()):
            # first copy last of the spectra that should be added
            sid = self.specIf.CopySpectrum(ids.pop(), addTo)

        # add all other spectra to the last spectrum
        for i in ids[1:]:
            try:
                hdtv.ui.msg("Adding " + str(i) + " to " + str(addTo))
                self.spectra.dict[addTo].Plus(self.spectra.dict[i])
            except KeyError:
                raise hdtv.cmdline.HDTVCommandError("Could not add " + str(i))
        self.spectra.dict[addTo].name = "sum"

        if args.normalize:
            norm_fac = len(ids)
            hdtv.ui.msg("Normalizing spectrum %s by 1/%d" % (addTo, norm_fac))
            self.spectra.dict[addTo].Multiply(1.0 / norm_fac)

    def SpectrumSub(self, args):
        """
        Subtract spectra (spec1 - spec2, ...)
        """
        # FIXME: Properly separate targetid, specid
        ids = hdtv.util.ID.ParseIds(
            [args.targetid] + args.specid, self.spectra, only_existent=False
        )

        if len(ids) == 0:
            hdtv.ui.warning("Nothing to do")
            return

        subFrom = ids[0]
        if subFrom not in list(self.spectra.dict.keys()):
            sid = self.specIf.CopySpectrum(ids.pop(), subFrom)

        for i in ids[1:]:
            try:
                hdtv.ui.msg("Subtracting " + str(i) + " from " + str(subFrom))
                self.spectra.dict[subFrom].Minus(self.spectra.dict[i])
            except KeyError:
                raise hdtv.cmdline.HDTVCommandError("Could not subtract " + str(i))
        self.spectra.dict[subFrom].name = "difference"

    def SpectrumMultiply(self, args):
        """
        Multiply spectrum
        """
        if args.specid is None:
            if self.spectra.activeID is not None:
                msg = (
                    "Using active spectrum %s for multiplication"
                    % self.spectra.activeID
                )
                hdtv.ui.msg(msg)
                ids = [self.spectra.activeID]
            else:
                hdtv.ui.msg("No active spectrum")
                ids = list()
        else:
            ids = hdtv.util.ID.ParseIds(args.specid, self.spectra)

        if len(ids) == 0:
            hdtv.ui.warning("Nothing to do")
            return

        for i in ids:
            if i in list(self.spectra.dict.keys()):
                hdtv.ui.msg("Multiplying " + str(i) + " with " + str(args.factor))
                self.spectra.dict[i].Multiply(args.factor)
            else:
                raise hdtv.cmdline.HDTVCommandError(
                    "Cannot multiply spectrum " + str(i) + " (Does not exist)"
                )

    def SpectrumRebin(self, args):
        """
        Rebin spectrum
        """
        if args.specid is None:
            if self.spectra.activeID is not None:
                msg = "Using active spectrum %s for rebinning" % self.spectra.activeID
                hdtv.ui.msg(msg)
                ids = [self.spectra.activeID]
            else:
                hdtv.ui.msg("No active spectrum")
                ids = list()
        else:
            ids = hdtv.util.ID.ParseIds(args.specid, self.spectra)

        if len(ids) == 0:
            hdtv.ui.warning("Nothing to do")
            return

        for i in ids:
            if i in list(self.spectra.dict.keys()):
                hdtv.ui.msg(
                    "Rebinning "
                    + str(i)
                    + " with "
                    + str(args.ngroup)
                    + " bins per new bin"
                )
                self.spectra.dict[i].Rebin(args.ngroup, calibrate=args.calibrate)
            else:
                raise hdtv.cmdline.HDTVCommandError(
                    "Cannot rebin spectrum " + str(i) + " (Does not exist)"
                )

    def SpectrumCalbin(self, args):
        """
        Calbin spectrum
        """
        if args.specid is None:
            if self.spectra.activeID is not None:
                msg = "Using active spectrum %s for rebinning" % self.spectra.activeID
                hdtv.ui.msg(msg)
                ids = [self.spectra.activeID]
            else:
                hdtv.ui.msg("No active spectrum")
                ids = list()
        else:
            ids = hdtv.util.ID.ParseIds(args.specid, self.spectra)

        use_tv_binning = hdtv.options.Get("spec.binning") == "tv"
        use_tv_binning |= args.tv_binning
        use_tv_binning ^= args.root_binning

        if use_tv_binning:
            hdtv.ui.debug(f"Calbinning using tv binning convention.")
        else:
            hdtv.ui.debug(f"Calbinning using ROOT binning convention.")

        if len(ids) == 0:
            hdtv.ui.warning("Nothing to do")
            return

        with hdtv.util.temp_seed(args.seed):
            for i in ids:
                if i in list(self.spectra.dict.keys()):
                    spec = self.spectra.dict[i]
                    sum_before = spec._hist.Integral()
                    spec.Calbin(
                        binsize=args.binsize,
                        spline_order=args.spline_order,
                        use_tv_binning=use_tv_binning,
                    )
                    if spec.cal:
                        self.spectra.caldict[spec.name] = spec.cal
                    if not args.deterministic:
                        spec.Poisson()
                    hdtv.ui.msg(f"Calbinning {i} with binsize={args.binsize}")
                    sum_after = spec._hist.Integral()
                    change = 100 * (1 - sum_after / sum_before)
                    hdtv.ui.debug(
                        "Calbin: "
                        f"Area before = {sum_before}. "
                        f"Area after = {sum_after}."
                    )
                    hdtv.ui.info(f"Total area changed by {change:f}%")
                else:
                    raise hdtv.cmdline.HDTVCommandError(
                        "Cannot rebin spectrum " + str(i) + " (Does not exist)"
                    )

    def SpectrumResample(self, args):
        """
        Rebin spectrum
        """
        if args.specid is None:
            if self.spectra.activeID is not None:
                msg = "Using active spectrum %s for resampling" % self.spectra.activeID
                hdtv.ui.msg(msg)
                ids = [self.spectra.activeID]
            else:
                hdtv.ui.msg("No active spectrum")
                ids = list()
        else:
            ids = hdtv.util.ID.ParseIds(args.specid, self.spectra)

        if len(ids) == 0:
            hdtv.ui.warning("Nothing to do")
            return

        with hdtv.util.temp_seed(args.seed):
            for i in ids:
                if i in list(self.spectra.dict.keys()):
                    spec = self.spectra.dict[i]
                    sum_before = spec._hist.Integral()
                    if args.distribution == "poisson":
                        spec.Poisson()
                    else:
                        raise hdtv.cmdline.HDTVCommandError(
                            f"Unknown probability distribution: {args.distribution}"
                        )
                    sum_after = spec._hist.Integral()
                    change = 100 * (1 - sum_after / sum_before)
                    hdtv.ui.debug(
                        f"Resample ({args.distribution}): "
                        f"Area before = {sum_before}. "
                        f"Area after = {sum_after}."
                    )
                    hdtv.ui.info(f"Total area changed by {change:f}%")
                else:
                    raise hdtv.cmdline.HDTVCommandError(
                        f"Cannot resample spectrum {i} (Does not exist)"
                    )

    def SpectrumHide(self, args):
        """
        Hides spectra
        """
        if len(args.specid) == 0:
            ids = list(self.spectra.dict.keys())
        else:
            ids = hdtv.util.ID.ParseIds(args.specid, self.spectra)

        self.spectra.HideObjects(ids)

    def SpectrumShow(self, args):
        """
        Shows spectra

        When inverse == True SpectrumShow behaves like SpectrumHide
        """
        if len(args.specid) == 0:
            ids = list(self.spectra.dict.keys())
        else:
            ids = hdtv.util.ID.ParseIds(args.specid, self.spectra)

        self.spectra.ShowObjects(ids)

    def SpectrumInfo(self, args):
        """
        Print info on spectrum objects
        """
        if len(args.specid) == 0:
            specids = ["active"]
        else:
            specids = args.specid
        ids = hdtv.util.ID.ParseIds(specids, self.spectra)
        s = ""
        for ID in ids:
            try:
                spec = self.spectra.dict[ID]
            except KeyError:
                s += "Spectrum %s: ID not found\n" % ID
                continue
            s += "Spectrum %s:\n" % ID
            s += hdtv.util.Indent(spec.info, "  ")
        hdtv.ui.msg(s, end="")

    def SpectrumUpdate(self, args):
        """
        Refresh spectra
        """
        if len(args.specid) == 0:
            specids = ["active"]
        else:
            specids = args.specid
        ids = hdtv.util.ID.ParseIds(specids, self.spectra)
        if len(ids) == 0:
            hdtv.ui.warning("Nothing to do")
            return
        self.spectra.RefreshObjects(ids)

    def SpectrumWrite(self, args):
        """
        Write Spectrum to File
        """
        # TODO: should accept somthing like "spec write all"
        filename = args.filename
        fmt = args.format
        if args.specid is None:
            ID = self.spectra.activeID
        else:
            ids = hdtv.util.ID.ParseIds(args.specid, self.spectra)
            if len(ids) != 1:
                raise ValueError("There is just one index possible here.")
            ID = ids[0]
        try:
            fname = hdtv.util.user_save_file(args.filename, args.force)
            if not fname:
                return
            self.spectra.dict[ID].WriteSpectrum(filename, fmt)
            hdtv.ui.msg(f"Wrote spectrum with id {ID} to file {filename}")
        except KeyError:
            hdtv.ui.warning("There is no spectrum with id: %s" % ID)

    def SpectrumName(self, args):
        """
        Give spectrum a name
        """
        if args.specid is None:
            ID = self.spectra.activeID
        else:
            ids = hdtv.util.ID.ParseIds(args.specid, self.spectra)

            if len(ids) == 0:
                hdtv.ui.warning("Nothing to do")
                return
            elif len(ids) > 1:
                hdtv.ui.warning("Can only rename one spectrum at a time")
                return

            ID = ids[0]

        spec = self.spectra.dict[ID]
        spec.name = args.name
        if spec.cal and not spec.cal.IsTrivial():
            self.spectra.caldict[args.name] = spec.cal
        hdtv.ui.msg(f"Renamed spectrum {ID} to '{args.name}'")

    def SpectrumNormalization(self, args):
        "Set normalization for spectrum"
        if args.specid is None:
            ids = [self.spectra.activeID]
        else:
            ids = hdtv.util.ID.ParseIds(args.specid, self.spectra)
            if len(ids) == 0:
                hdtv.ui.warning("Nothing to do")
                return

        for ID in ids:
            try:
                self.spectra.dict[ID].norm = args.norm
            except KeyError:
                raise hdtv.cmdline.HDTVCommandError(
                    "There is no spectrum with id: %s" % ID
                )

    def RandomModelCompleter(self, text, args=None):
        """
        Helper function for SpectrumResample
        """
        return hdtv.util.GetCompleteOptions(text, iter(["poisson"]))


# plugin initialisation
import __main__

spec_interface = SpecInterface(__main__.spectra)
hdtv.cmdline.RegisterInteractive("s", spec_interface)



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA


try:
    import curses
    import io

    def get_term_width():
        # We call curses.setupterm() every time, as the terminal width
        # may have changed
        try:
            curses.setupterm()
            return curses.tigetnum("cols")
        except BaseException:
            return 80

except ImportError:
    # On platforms where we cannot load the curses library, we assume a
    # terminal width of 80 characters
    def get_term_width():
        return 80


def tabformat(cells, **kwargs):
    """
    Prints a number of cells in a layout similar to that used
    by the ``ls'' program. Expects a list of cell contents.
    Optional keyword arguments are colsepwidth, the spacing between
    two columns (defaults to 2), and tabwidth, the requested width
    of the table (defaults to width of terminal, or 80 if that
    cannot be read).
    """
    # Read keyword arguments, replace with defaults if required
    if "colsepwidth" in kwargs:
        col_sep_width = kwargs["colsepwidth"]
    else:
        col_sep_width = 2

    if "tabwidth" in kwargs:
        tabwidth = kwargs["tabwidth"]
    else:
        tabwidth = get_term_width()

    # If there are no cells, there is nothing to do...
    n_cells = len(cells)
    if n_cells == 0:
        return

    # Gather some statistics about the cells
    cell_widths = [len(cell) for cell in cells]
    min_cell_width = min(cell_widths)
    max_cell_width = max(cell_widths)

    # We use a rather simple-minded algorithm here: we know that at least one
    # column is as wide as the widest cell, while each column is at least as
    # wide as the smallest cell. This allows us to obtain an upper bound for
    # the number of columns, and thus a lower bound for the number of rows.
    # We then try to actually layout the table; if that fails, we increase the
    # decrease the number of rows and try again.
    #
    # Note that we need to try each number of rows, even if the number of
    # columns does not change, since changing the number of rows influences
    # the distribution of cells into columns, and thus the column widths.

    # Check if a two-column table *may* be possible. If not, we know that
    # we need a one-column table.
    if (max_cell_width + min_cell_width + col_sep_width) > tabwidth:
        n_cols = 1
        n_rows = n_cells
    else:
        # Calculate an upper bound for the number of columns. Note that the space
        # between two columns takes some width as well.
        n_cols = (tabwidth - max_cell_width) // (min_cell_width + col_sep_width) + 1

        # Calculate the corresponding number of rows as
        # n_rows = ceil(n_cells / n_cols), for integers
        n_rows = (n_cells - 1) // n_cols + 1

        while True:
            # Calculate the minimal number of columns for the given number of rows
            # n_cols = ceil(n_cells / n_rows), for integers
            n_cols = (n_cells - 1) // n_rows + 1

            # Try with n_cols columns, and calculate the table width.
            tbl_width = 0

            for i in range(0, n_cols):
                tbl_width += max(cell_widths[i * n_rows : (i + 1) * n_rows])
            # If the table is small enough, end the loop...
            if tbl_width + (n_cols - 1) * col_sep_width <= tabwidth:
                break
            # ...otherwise, increase the number of rows by one and try again
            n_rows += 1

    # Now produce the actual output
    # Calculate the individual column widths
    col_widths = []
    for i in range(0, n_cols):
        col_widths.append(max(cell_widths[i * n_rows : (i + 1) * n_rows]))

    # Distribute cells over rows, aloowing the table to be printed
    # row-by-row (note that successive cells go *below* each other)
    rows = [[] for i in range(0, n_rows)]
    for i in range(0, n_cells):
        rows[i % n_rows].append(cells[i])

    # Fill up rows with empty cells
    for i in range(0, n_rows):
        if len(rows[i]) < n_cols:
            rows[i].append("")

    # Produce the appropriate format string for output
    fmtstr = (" " * col_sep_width).join(["%%-%ds" % w for w in col_widths])

    # Output the table, row by row
    for i in range(0, n_rows):
        print(fmtstr % tuple(rows[i]))
